@@ -8,7 +8,7 @@ Licensed under the terms contained in the file 'COPYING' in the
 source directory.
 
 */
-
+#define BIG_CHAIN 5
 
 #include <stdlib.h>
 #include <stdio.h>
@@ -31,6 +31,11 @@ source directory.
 #include "boot.h"
 #include "loader.h"
 
+/* the number of sectors between the 15M memory hole and 1M
+   this is the max. for a bzImage kernel + initrd unless "large-memory"
+   or "mem=XXX" raises the limit  			*/
+#define HIGH_SECTORS	((15-1)*1024*1024/SECTOR_SIZE) 
+#define HIGH_4M		(3*1024*1024/SECTOR_SIZE)
 
 static GEOMETRY geo;
 static struct stat st;
@@ -48,7 +53,7 @@ void boot_image(char *spec,IMAGE_DESCR *descr)
     BOOT_SECTOR buff;
     SETUP_HDR hdr;
     char *initrd;
-    int setup,fd,sectors;
+    int setup,fd,sectors,hi_sectors=MAX_KERNEL_SECS*4;
     int modern_kernel;
 
     if (verbose > 0) {
@@ -60,11 +65,7 @@ void boot_image(char *spec,IMAGE_DESCR *descr)
     if (fstat(fd,&st) < 0) die("fstat %s: %s",spec,strerror(errno));
     if (read(fd,(char *) &buff,SECTOR_SIZE) != SECTOR_SIZE)
 	die("read %s: %s",spec,strerror(errno));
-#ifdef LCF_VARSETUP
     setup = buff.sector[VSS_NUM] ? buff.sector[VSS_NUM] : SETUPSECS;
-#else
-    setup = SETUPSECS;
-#endif
     if (read(fd,(char *) &hdr,sizeof(hdr)) != sizeof(hdr))
 	die("read %s: %s",spec,strerror(errno));
     modern_kernel = !strncmp(hdr.signature,NEW_HDR_SIG,4) && hdr.version >=
@@ -73,16 +74,16 @@ void boot_image(char *spec,IMAGE_DESCR *descr)
     if (verbose > 1)
 	printf("Setup length is %d sector%s.\n",setup,setup == 1 ? "" : "s");
     map_add(&geo,0,(st.st_size+SECTOR_SIZE-1)/SECTOR_SIZE);
-    sectors = map_end_section(&descr->start,setup+SPECIAL_SECTORS);
+    sectors = map_end_section(&descr->start,setup+SPECIAL_SECTORS+SPECIAL_BOOTSECT);
     if (!modern_kernel || !(hdr.flags & LFLAG_HIGH))
 	check_size(spec,setup,sectors);
     else {
 	if (hdr.start % PAGE_SIZE)
 	    die("Can't load kernel at mis-aligned address 0x%08lx\n",hdr.start);
-#if 0
-	descr->start_page = hdr.start/PAGE_SIZE; /* load kernel high */
-#endif
 	descr->flags |= FLAG_LOADHI;	/* load kernel high */
+	hi_sectors = sectors - setup;	/* number of sectors loaded high */
+	hi_sectors *= 3;		/* account for decompression */
+	if (hi_sectors < HIGH_4M) hi_sectors = HIGH_4M;
     }
     geo_close(&geo);
     if (verbose > 1)
@@ -108,6 +109,28 @@ void boot_image(char *spec,IMAGE_DESCR *descr)
 	if (verbose > 1)
 	    printf("RAM disk: %d sector%s.\n",sectors,sectors == 1 ?  "" :
 	      "s");
+	if (hi_sectors + sectors > HIGH_SECTORS
+#ifndef LCF_INITRDLOW
+	    && !cfg_get_flag(cf_options,"large-memory")
+#endif
+	    ) {
+		descr->flags |= FLAG_TOOBIG;
+		if (!nowarn)
+		fprintf(errstd, "Warning: The initial RAM disk is too big to fit between %s and\n"
+				"   the 15M-16M memory hole."
+#ifndef LCF_INITRDLOW
+# if 0
+							"  If your BIOS supports memory moves above 16M,\n"
+				"   then you may specify \"large-memory\" in the configuration file\n"
+				"   (/etc/lilo.conf)."
+# else
+							"  It will be loaded in the highest memory as\n"
+				"   though the configuration file specified \"large-memory\" and it will\n"
+				"   be assumed that the BIOS supports memory moves above 16M."
+# endif
+#endif
+				"\n", hi_sectors ? "the kernel" : "1M");
+	}
 	geo_close(&geo);
     }
 }
@@ -252,7 +275,7 @@ char *boot_mbr(const char *boot, int table)
 	k = dev & ~mask;
 	k = !table ? 1 : k>=1 && k<=4;
 	if (k) {
-	    dev_open(&d, dev&mask, -1);	/* bypass any open */
+	    dev_open(&d, dev&mask, O_BYPASS);	/* bypass any open */
 	    npart = stralloc(d.name); 
 	    dev_close(&d);
 	}
@@ -284,10 +307,16 @@ void boot_other(char *loader,char *boot,char *part,IMAGE_DESCR *descr)
     int l_fd;
 #endif
     unsigned short magic;
+#ifdef BIG_CHAIN
+    BOOT_SECTOR buff[BIG_CHAIN];
+    BOOT_SECTOR zbuff;
+    int mapped;
+#else
     BOOT_SECTOR buff[SETUPSECS-1];
+#endif
     struct stat st;
     char *pos;
-    int i;
+    int i, code;
     int letter = 0;
 
     if (!loader) loader = DFL_CHAIN;
@@ -357,17 +386,49 @@ void boot_other(char *loader,char *boot,char *part,IMAGE_DESCR *descr)
 	if (magic != BOOT_SIGNATURE)
 	    die("First sector of %s doesn't have a valid boot signature",boot);
     }
-    if (cfg_get_flag(cf_other,"master-boot")) {
+
+/* process the 'master-boot' or 'boot-as' options */
+    i = cfg_get_flag(cf_other,"master-boot");
+    pos = cfg_get_strg(cf_other,"boot-as");
+    if (i && pos) die("'master-boot' and 'boot-as' are mutually exclusive 'other=' options");
+    if (!i && !pos) {
+	i = cfg_get_flag(cf_options,"master-boot");
+	pos = cfg_get_strg(cf_options,"boot-as");
+	if (i && pos) die("'master-boot' and 'boot-as' are mutually exclusive global options");
+    }
+    if (i) code = -1;	/* master-boot in use */
+    else if (pos) {
+	code = to_number(pos);
+	if (code >= 80 && code <= 89) {
+	    /* convert to 0x80 to 0x89 */
+	    fprintf(errstd,"Warning: Radix error, 'boot-as=%d' taken to mean 'boot-as=0x%x'\n",
+	    			code, code+0x30);
+	    code += 0x30;
+	}
+	if ( !((code>=0 && code<=3) || (code>=0x80 && code<=DEV_MASK)) )
+	    die("Illegal BIOS device code specified in 'boot-as=0x%02x'", code);
+    }
+    else code = -2;
+    
+    if (code != -2) {
 	curr_drv_map += 2;	/* add 2 spaces */
 	if (curr_drv_map >= DRVMAP_SIZE)
 	    cfg_error("Too many drive mappings (more than %d)",DRVMAP_SIZE);
-	if (verbose > 1)
-	    printf("  Swapping BIOS boot drive with 0/0x80, as needed\n");
+	if (verbose > 1) {
+	    char *s, n[8];
+	    if (code==-1) s = "0/0x80";
+	    else sprintf((s=n),"0x%02x", code);
+	    printf("  Swapping BIOS boot drive with %s, as needed\n", s);
+	}
 	for (i=curr_drv_map-1; i>1; i--) drv_map[i] = drv_map[i-2];
-	drv_map[1] = drv_map[0] = 0xFFFF;	/* reserve 2 slots */
+	drv_map[0] = 0xFFFF;	/* reserve 2 slots */
+	drv_map[1] = code<<8 | 0xFF;
     }
 
     memset(buff,0,sizeof(buff));
+#ifdef BIG_CHAIN
+    zbuff = buff[0];	/* zero out zbuff */
+#endif
 #ifndef LCF_BUILTIN
     if ((l_fd = open(loader,O_RDONLY)) < 0)
 	die("open %s: %s",loader,strerror(errno));
@@ -391,7 +452,7 @@ void boot_other(char *loader,char *boot,char *part,IMAGE_DESCR *descr)
     else {
 	if ((p_fd = open(part,O_RDONLY)) < 0)
 	    die("open %s: %s",part,strerror(errno));
-	if (lseek(p_fd,(long) PART_TABLE_OFFSET,0) < 0)
+	if (lseek(p_fd,(long) PART_TABLE_OFFSET,SEEK_SET) < 0)
 	    die("lseek %s: %s",part,strerror(errno));
 	if (read(p_fd,(char *) buff[0].par_c.ptable,PART_TABLE_SIZE) !=
 	  PART_TABLE_SIZE)
@@ -430,19 +491,37 @@ void boot_other(char *loader,char *boot,char *part,IMAGE_DESCR *descr)
     pos = (char *) buff+buff[0].par_c.drvmap;
     memcpy(pos,drv_map,sizeof(drv_map));
     memcpy(pos+sizeof(drv_map),prt_map,sizeof(prt_map)-2);
-    map_add_zero();
-#if 0
-    for (i = 0; i < SETUPSECS-1; i++) map_add_sector(&buff[i]);
-#else
+
     size = (size + SECTOR_SIZE - 1) / SECTOR_SIZE;
-    for (i = 0; i < size; i++) map_add_sector(&buff[i]);
-    for (; i < SETUPSECS-1; i++) map_add_zero();
+#ifndef BIG_CHAIN
+    map_add_zero();
+#else
+    if (size > SETUPSECS-1) {
+	zbuff.sector[VSS_NUM] = mapped = size+1;
+	map_add_sector(zbuff.sector);
+    }
+    else {
+	map_add_zero();
+	mapped = SETUPSECS;
+    }
 #endif
+    for (i = 0; i < size; i++) map_add_sector(&buff[i]);
+    for (i = size; i < SETUPSECS-1; i++) map_add_zero();
     map_add(&geo,0,1);
+
+#ifndef BIG_CHAIN
     (void) map_end_section(&descr->start,SETUPSECS+SPECIAL_SECTORS);
 	/* size is known */
     geo_close(&geo);
     if (verbose > 1)
 	printf("Mapped %d (%d+1+1) sectors.\n",
 		SETUPSECS+SPECIAL_SECTORS, SETUPSECS);
+#else
+    (void) map_end_section(&descr->start, mapped+SPECIAL_SECTORS+SPECIAL_BOOTSECT);
+	/* size is known */
+    geo_close(&geo);
+    if (verbose > 1)
+	printf("Mapped %d (%d+1+1) sectors.\n",
+		mapped+SPECIAL_SECTORS, mapped);
+#endif
 }
