@@ -62,6 +62,7 @@ static char secondary_map[SECTOR_SIZE];
 static unsigned char table[SECTOR_SIZE];	/* keytable & params */
 static DEVICE dev;
 static char *boot_devnam,*map_name;
+static int use_dev_close = 0;
 static int fd;
 static int image_base = 0,image = 0;
 static char temp_map[PATH_MAX+1];
@@ -94,11 +95,11 @@ BUILTIN_FILE *select_loader(void)
     } 
     else if (strstr(install,"text")) loader = &Second;  /* text interface */
     else if (strstr(install,"menu")) loader = &Third;
-    else if (bitmap) loader = &Bitmap;   /* menu interface (MDA, EGA, VGA) */
+    else if (strstr(install,"bmp") || bitmap) loader = &Bitmap;   /* menu interface (MDA, EGA, VGA) */
         
     adapter = get_video();	/* from probe.c */
     if (!nowarn) {
-	if (adapter <= VIDEO_UNKNOWN)
+	if (adapter <= VIDEO_UNKNOWN && (verbose>=2 || loader==&Bitmap))
 	    fprintf(errstd, "Warning: Unable to determine video adapter in use in the present system.\n");
 	if (loader == &Third && adapter == VIDEO_CGA)
 	    fprintf(errstd, "Warning: Video adapter (CGA) is incompatible with the boot loader selected for\n"
@@ -131,6 +132,7 @@ static char *disallow[] = {
     while (*dis) {
 	here = strstr(options, *dis);
 	if (here) {
+	    if (here[3] == '=') error=2;
 	    if (here[3] != '=' && (here[4]==' ' || here[4]==0) ) error=2;
 #if !MEM_OKAY
 	    if (*dis == disallow[0] && here[4]>='0' && here[4]<='9')
@@ -454,14 +456,7 @@ static void open_bsect(char *boot_dev)
 	printf("Reading boot sector from %s\n",boot_dev ? boot_dev :
 	  "current root.");
     boot_devnam = boot_dev;
-    if (boot_dev) {
-	if ((fd = open(boot_dev,O_RDWR)) < 0)
-	    die("open %s: %s",boot_dev,strerror(errno));
-	if (fstat(fd,&st) < 0) die("stat %s: %s",boot_dev,strerror(errno));
-	if (!S_ISBLK(st.st_mode)) boot_dev_nr = 0;
-	else boot_dev_nr = st.st_rdev;
-    }
-    else {
+    if (!boot_dev || !strcmp(boot_dev, "/") ) {
 	if (stat("/",&st) < 0) pdie("stat /");
 #if 0
 	if ((st.st_dev & PART_MASK) > PART_MAX)
@@ -473,8 +468,18 @@ static void open_bsect(char *boot_dev)
 	      (int)st.st_dev);
 	fd = dev_open(&dev,boot_dev_nr = st.st_dev,O_RDWR);
 	boot_devnam = dev.name;
+	use_dev_close = 1;
+    }
+    else {
+	if ((fd = open(boot_dev,O_RDWR)) < 0)
+	    die("open %s: %s",boot_dev,strerror(errno));
+	if (fstat(fd,&st) < 0) die("stat %s: %s",boot_dev,strerror(errno));
+	if (!S_ISBLK(st.st_mode)) boot_dev_nr = 0;
+	else boot_dev_nr = st.st_rdev;
     }
 /* new code to get boot device code */
+/* plus geo_open will trigger VolumeMgmt (pf_hard_disk_scan) before the
+   boot sector is actually read; timing is very important */
 {
     GEOMETRY geo;
     geo_open(&geo, boot_devnam, O_RDONLY);
@@ -488,7 +493,18 @@ static void open_bsect(char *boot_dev)
     if (read(fd,(char *) &bsect,SECTOR_SIZE) != SECTOR_SIZE)
 	die("read %s: %s",boot_dev ? boot_dev : dev.name,strerror(errno));
     bsect_orig = bsect;
-    ireloc = part_nowrite(boot_dev);
+    ireloc = part_nowrite(boot_devnam);
+    if (ireloc == PTW_DOS + PTW_NTFS)  ireloc = PTW_DOS;
+/* check for override (-F) command line flag */
+    if (ireloc>PTW_DOS && force_fs) {
+#if 0
+	fprintf(errstd, "WARNING: '-F' override used. Filesystem on  %s  may be destroyed.\n", boot_devnam);
+#else
+	fprintf(stderr, "WARNING: '-F' override used. Filesystem on  %s  may be destroyed.\n", boot_devnam);
+	if (!yesno("\nProceed? ",0)) exit(0);
+#endif
+	ireloc=PTW_OKAY;
+    }
 }
 
 
@@ -587,10 +603,10 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 	die("Map %s is not a regular file.",map_file);
     open_bsect(boot_dev);
     part_verify(boot_dev_nr,1);
-    if (ireloc>1) {
+    if (ireloc>PTW_DOS) {
     	die("Filesystem would be destroyed by LILO boot sector: %s", boot_dev);
     }
-    else if (ireloc==1) {
+    else if (ireloc==PTW_DOS) {
 	if (!nowarn) fprintf(errstd,"Warning: boot record relocation beyond BPB is necessary: %s\n", boot_dev);
     }
 
@@ -632,7 +648,12 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 		required, because this map is built in memory */
 	here2 = map_insert_data (loader->data, loader->size);
 	memcpy(&param2,loader->data,sizeof(param2));
+#ifdef LCF_FIRST6
+	/* write just the 4 bytes (sa6==2) */
+	sectors = map_write((SECTOR_ADDR*)secondary_map, (SECTOR_SIZE-4)/sizeof(SECTOR_ADDR)-2, 1, 2);
+#else
 	sectors = map_write((SECTOR_ADDR*)secondary_map, (SECTOR_SIZE-4)/sizeof(SECTOR_ADDR)-2, 1);
+#endif
 	memcpy(secondary_map+SECTOR_SIZE-4, EX_MAG_STRING, 4);
 
 	/* fill in full size of secondary boot loader in paragraphs */
@@ -662,10 +683,24 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 
 	map_begin_section();
 	map_add_sector(secondary_map);
+#ifdef LCF_FIRST6
+	/* write out six byte address */
+	(void) map_write(&bsect.par_1.secondary,1,0,1);
+#else
 	(void) map_write(&bsect.par_1.secondary,1,0);
-	if (bios_boot == bios_map)
+#endif
+
+/* if the state of the BIOS is DL_GOOD, always mark when boot==map
+   if the state of the BIOS is < DL_GOOD, never mark */
+   
+	if ( bios_boot == bios_map  &&
+		(bios_passes_dl == DL_GOOD
+/***		 || (do_md_install && !(do_md_install==X_MBR_ONLY )) 	***/
+		  ) )
 	    bsect.par_1.prompt |= FLAG_MAP_ON_BOOT;
-	if (bios_map & 0x80)	/* hard disk? */
+	if (bios_boot!=bios_map && !nowarn)
+	    fprintf(errstd, "Warning: The boot sector and map file are on different disks.\n");
+	if ( (bios_map & 0x80) && !do_md_install) /* hard disk & not raid master */
 	    bsect.par_1.map_serial_no = serial_no[bios_map - 0x80];
 	if (verbose>=2)
 	    printf("bios_boot = 0x%02X  bios_map = 0x%02X  map==boot = %d  map S/N: %08lX\n",
@@ -682,7 +717,7 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
     message = cfg_get_strg(cf_options,"message");
     scheme = cfg_get_strg(cf_options,"bitmap");
     if (message && scheme) die("'bitmap' and 'message' are mutually exclusive");
-    bsect.par_1.msg_len = 0;
+    param2.msg_len = 0;
     bitmap = (loader==&Bitmap);
     if (bitmap) {
 	message = scheme;
@@ -731,7 +766,7 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 	i = bitmap ? MAX_KERNEL_SECS*SECTOR_SIZE : MAX_MESSAGE;
 	if (st.st_size > i)
 	    die("%s is too big (> %d bytes)",message,i);
-	bsect.par_1.msg_len = bitmap ? (st.st_size+15)/16 : st.st_size;
+	param2.msg_len = bitmap ? (st.st_size+15)/16 : st.st_size;
 	map_begin_section();
 #ifndef LCF_UNIFY
 	map_add(&geo,0,((st.st_size)+SECTOR_SIZE-1)/SECTOR_SIZE);
@@ -744,14 +779,18 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 	    		sectors,sectors == 1 ?  "" : "s");
 	geo_close(&geo);
     }
+
+    if (cfg_get_flag(cf_options,"el-torito-bootable-cd"))
+    param2.flag2 |= FLAG2_EL_TORITO;
+    
     serial = cfg_get_strg(cf_options,"serial");
     if (serial) {
     	if (!(stage_flags & STAGE_FLAG_SERIAL))
     	    die("Serial line not supported by boot loader");
 	if (!*serial || !(this = strchr(coms,*serial)))
 	    die("Invalid serial port in \"%s\" (should be 0-3)",serial);
-	else bsect.par_1.port = (this-coms)+1;
-	bsect.par_1.ser_param = SER_DFL_PRM;
+	else param2.port = (this-coms)+1;
+	param2.ser_param = SER_DFL_PRM;
 	if (serial[1]) {
 	    if (serial[1] != ',')
 		die("Serial syntax is <port>[,<bps>[<parity>[<bits>]]]");
@@ -762,30 +801,30 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 	        walk += i+1;
 	    }
 	    if (!*walk) die("Unsupported baud rate");
-	    bsect.par_1.ser_param &= ~0xE4;
+	    param2.ser_param &= ~0xE4;
 	    if (speed==16) speed -= 6;  /* convert 56000 to 57600 */
-	    bsect.par_1.ser_param |= ((speed<<5) | (speed>>1)) & 0xE4;
+	    param2.ser_param |= ((speed<<5) | (speed>>1)) & 0xE4;
 	    serial += i+2;
 	/* check for parity specified */
 	    if (*serial) {
 		if (!(this = strchr(parity,*serial)))
-		    die("Valid parity values are N, O and E");
+		    die("Serial speed = %s; valid parity values are N, O and E", walk);
 		i = (int)(this-parity)>>1;
 		if (i==2) i++; /* N=00, O=01, E=11 */
-		bsect.par_1.ser_param &= ~(i&1); /* 7 bits if parity specified */
-		bsect.par_1.ser_param |= i<<3;   /* specify parity */
+		param2.ser_param &= ~(i&1); /* 7 bits if parity specified */
+		param2.ser_param |= i<<3;   /* specify parity */
 	/* check if number of bits is there */
 		if (serial[1]) {
 		    if (serial[1] != '7' && serial[1] != '8')
 			die("Only 7 or 8 bits supported");
-		    if (serial[1]=='7')	bsect.par_1.ser_param &= 0xFE;
-		    else bsect.par_1.ser_param |= 0x01;
+		    if (serial[1]=='7')	param2.ser_param &= 0xFE;
+		    else param2.ser_param |= 0x01;
 		    
 		    if (serial[2]) die("Synax error in SERIAL");
 		}
 	    }
 	    if (verbose>=4) printf("Serial Param = 0x%02X\n", 
-	                                        (int)bsect.par_1.ser_param);
+	                                        (int)param2.ser_param);
 	}
 	if (delay < 20 && !cfg_get_flag(cf_options,"prompt")) {
 	    fprintf(errstd,"Setting DELAY to 20 (2 seconds)\n");
@@ -797,6 +836,17 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
     	fprintf(errstd,"CAUTION:  boot-time BIOS data will not be saved\n");
     	bsect.par_1.prompt |= FLAG_NOBD;
     }
+    if (!fetch() && (bios_map &
+#if VERSION_MINOR>=50
+    			bios_boot &	/* if 'boot=/dev/fd0', force write */
+#endif
+    				0x80)) {
+	bsect.par_1.prompt |= FLAG_BD_OKAY;
+	if (verbose>=2) printf("BIOS data check was okay on the last boot\n");
+    }
+    else {
+	if (verbose>=2) printf("BIOS data check will include auto-suppress check\n");
+    }
     if (cfg_get_flag(cf_options,"large-memory")) {
 #ifndef LCF_INITRDLOW
 	bsect.par_1.prompt |= FLAG_LARGEMEM;
@@ -816,10 +866,10 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 #define tick(x) ((x)*4000/2197)
 #endif
     if (tick(delay) > 0xffff) die("Maximum delay is 59:59 (3599.5secs).");
-	else bsect.par_1.delay = tick(delay);
-    if (timeout == -1) bsect.par_1.timeout = 0xffff;
+	else param2.delay = tick(delay);
+    if (timeout == -1) param2.timeout = 0xffff;
     else if (tick(timeout) >= 0xffff) die("Maximum timeout is 59:59 (3599.5secs).");
-	else bsect.par_1.timeout = tick(timeout);
+	else param2.timeout = tick(timeout);
 
 /* keytable & parameter area setup */
 
@@ -876,7 +926,7 @@ printf("MAX_IMAGES = %d\n", MAX_IMAGES);
 #if 0
     map_begin_section();
     map_add_sector(table);
-    (void) map_write(&bsect.par_1.keytab,1,0);
+    (void) map_write(&param2.keytab,1,0);
 #endif
     memset(&descrs,0,SECTOR_SIZE*MAX_DESCR_SECS);
     if (cfg_get_strg(cf_options,"default")) image = image_base = 1;
@@ -1149,6 +1199,9 @@ static void bsect_done(char *name,IMAGE_DESCR *descr)
 		(this>=0 && (descrs.d.descr[this].flags & FLAG_VMDEFAULT)) )
 	    printf(" @");
 #endif
+	if (descrs.d.descr[this_image].flags & FLAG_TOOBIG ||
+		(this>=0 && (descrs.d.descr[this].flags & FLAG_TOOBIG)) )
+	    printf(" ?");
 	if (this_image && this) putchar('\n');
 	else printf(" *\n");
     }
@@ -1166,7 +1219,8 @@ static void bsect_done(char *name,IMAGE_DESCR *descr)
 
 int bsect_number(void)
 {
-    return image_base ? 0 : image;
+ /* -1 means default= did not exist */
+    return image_base ? -1 : image;
 }
 
 
@@ -1203,12 +1257,13 @@ void check_fallback(void)
 void bsect_update(char *backup_file, int force_backup, int pass)
 {
     BOOT_SECTOR bsect_wr;
-static int timestamp = 0;
+    long temp;
+static long timestamp = 0;
 
-    if (pass>=0 && !timestamp) {
-        timestamp = make_backup(backup_file, force_backup, &bsect_orig,
+    if (pass>=0) {
+        temp = make_backup(backup_file, force_backup, &bsect_orig,
     						boot_dev_nr, "boot sector");
-        if (timestamp) bsect.par_1.timestamp = timestamp;
+        if (temp && !timestamp) bsect.par_1.timestamp = timestamp = temp;
     }
 
 #ifndef LCF_UNIFY
@@ -1223,13 +1278,17 @@ static int timestamp = 0;
 	((long*)table)[SECTOR_SIZE/sizeof(long)-2] = crc32(table, SECTOR_SIZE-2*sizeof(long), CRC_POLY1);
 	map_begin_section();
 	map_add_sector(table);
-	(void) map_write(&bsect.par_1.keytab,1,0);
-
+#ifdef LCF_FIRST6
+	/* still use 5 byte address */
+	(void) map_write(&param2.keytab,1,0,0);
+#else
+	(void) map_write(&param2.keytab,1,0);
+#endif
 	map_close(&param2, here2);
     }	/* if (pass<1) ...	*/
 
 if (pass>=0) {
-    if (lseek(fd,0,0) < 0)
+    if (lseek(fd,0,SEEK_SET) < 0)
 	die("lseek %s: %s",
 		boot_devnam ? boot_devnam : dev.name,
 		strerror(errno));
@@ -1285,13 +1344,12 @@ if (pass>=0) {
 
 } /* if (pass>=0) ... */
 
-    if (!boot_devnam) dev_close(&dev);
+    if (use_dev_close) dev_close(&dev);
     else if (close(fd) < 0) {
 	    unbootable();
 	    die("close %s: %s",boot_devnam,strerror(errno));
 	}
 
-
 #if 0
     if (pass==0) {
 #else
@@ -1304,7 +1362,9 @@ if (pass>=0) {
 	    die("rename %s %s: %s",temp_map,map_name,strerror(errno));
 	}
     }
-    (void) sync();
+/*  (void) sync();   Now handled in lilo.c (atexit(sync)) */
+	if (verbose>=6) printf("End  bsect_update\n");
+	fflush(stdout);
 }
 
 
@@ -1314,7 +1374,7 @@ void bsect_cancel(void)
     map_descrs(&descrs, bsect.par_1.descr, &bsect.par_1.dflcmd);
 #endif
     map_close(NULL,0);
-    if (boot_devnam) (void) close(fd);
+    if (!use_dev_close) (void) close(fd);
     else dev_close(&dev);
     temp_unregister(temp_map);
     if (verbose<9) (void) remove(temp_map);
@@ -1422,12 +1482,12 @@ void bsect_uninstall(char *boot_dev,char *backup_file,int validate)
     if (verbose > 0) printf("Reading old boot sector.\n");
     if (read(bck_file,(char *) &bsect,PART_TABLE_OFFSET) != PART_TABLE_OFFSET)
 	die("read %s: %s",backup_file,strerror(errno));
-    if (lseek(fd,0,0) < 0)
+    if (lseek(fd,0,SEEK_SET) < 0)
 	die("lseek %s: %s",boot_dev ? boot_dev : dev.name,strerror(errno));
     if (verbose > 0) printf("Restoring old boot sector.\n");
     if (write(fd,(char *) &bsect,PART_TABLE_OFFSET) != PART_TABLE_OFFSET)
 	die("write %s: %s",boot_dev ? boot_dev : dev.name,strerror(errno));
-    if (!boot_devnam) dev_close(&dev);
+    if (use_dev_close) dev_close(&dev);
     else if (close(fd) < 0) {
 	    unbootable();
 	    die("close %s: %s",boot_devnam,strerror(errno));
@@ -1440,6 +1500,7 @@ void bsect_raid_update(char *boot_dev, unsigned long raid_offset,
 	char *backup_file, int force_backup, int pass, int mask)
 {
     BOOT_SECTOR bsect_save;
+    int bios;
     int prompt = bsect.par_1.prompt;
 
     if (pass<0) bsect_update(backup_file, force_backup, pass);
@@ -1449,12 +1510,23 @@ void bsect_raid_update(char *boot_dev, unsigned long raid_offset,
         open_bsect(boot_dev);
         memcpy(&bsect, &bsect_save, MAX_BOOT_SIZE);	/* update the subject boot sector */
         bsect.par_1.raid_offset = raid_offset;	/* put in the new partition offset */
-#if 0
-        bsect.par_1.prompt &= FLAG_SAVE;	/* clear all RAID flags */
-#else
         bsect.par_1.prompt &= mask;		/* clear all RAID flags */
-#endif
         bsect.par_1.prompt |= raid_flags;	/* update the raid flags */
+
+        bios = (raid_flags&FLAG_RAID_DEFEAT) ? bios_map : bios_boot;
+        if (verbose>=2) printf("Using s/n from device 0x%02X\n", bios);
+	bsect.par_1.map_serial_no = serial_no[bios - 0x80];
+
+#ifdef LCF_FIRST6
+/* lines added 22.5.7 */
+	((SECTOR_ADDR6*)&bsect.par_1.secondary)->device = bios;
+#else
+/* lines added 22.5.6 */
+	bsect.par_1.secondary.device &= ~DEV_MASK;
+	bsect.par_1.secondary.device |= bios;
+#endif
+/* ************************************ */
+
         *(unsigned short *) &bsect.sector[BOOT_SIG_OFFSET] = BOOT_SIGNATURE;
     }
     
