@@ -58,80 +58,84 @@ CONFIG cf_identify[] = {
   { cft_end,  NULL,		NULL,		NULL,NULL }};
 
 CONFIG cf_options[] = {
+  { cft_strg, "append",		NULL,		NULL,NULL },
+  { cft_strg, "backup",		NULL,		NULL,NULL },
+  { cft_strg, "bios-passes-dl",	NULL,		NULL,NULL },
+  { cft_strg, "bitmap",		NULL,		NULL,NULL },
+  { cft_strg, "bmp-colors",	NULL,		NULL,NULL },
+  { cft_strg, "bmp-table",	NULL,		NULL,NULL },
+  { cft_strg, "bmp-timer",	NULL,		NULL,NULL },
   { cft_strg, "boot",		NULL,		NULL,NULL },
-  { cft_strg, RAID_EXTRA_BOOT,	NULL,		NULL,NULL },
+  { cft_strg, "boot-as",	NULL,		NULL,NULL },
+  { cft_flag, "change-rules",	do_cr,		NULL,NULL },
   { cft_flag, "compact",	NULL,		NULL,NULL },
   { cft_strg, "default",	NULL,		NULL,NULL },
-  { cft_strg, "vmdefault",	NULL,		NULL,NULL },
   { cft_strg, "delay",		NULL,		NULL,NULL },
-  { cft_strg, "timeout",	NULL,		NULL,NULL },
-  { cft_flag, "prompt",		NULL,		NULL,NULL },
-  { cft_flag, "suppress-boot-time-BIOS-data",	NULL,	NULL,NULL },
-  { cft_flag, "linear",     	NULL,		NULL,NULL },
-  { cft_flag, "lba32",          NULL,           NULL,NULL }, 
-  { cft_flag, "geometric",      NULL,           NULL,NULL }, 
-  { cft_flag, "fix-table",     	NULL,		NULL,NULL },
-  { cft_flag, "ignore-table",   NULL,		NULL,NULL },
-  { cft_strg, "install",	NULL,		NULL,NULL },
-  { cft_strg, "disktab",	NULL,		NULL,NULL },
   { cft_strg, "disk",		do_disk,	NULL,NULL },
-  { cft_strg, "map",		NULL,		NULL,NULL },
-  { cft_strg, "message",	NULL,		NULL,NULL },
-  { cft_strg, "verbose",	NULL,		NULL,NULL },
-  { cft_strg, "backup",		NULL,		NULL,NULL },
-  { cft_strg, "force-backup",	NULL,		NULL,NULL },
-  { cft_strg, "serial",		NULL,		NULL,NULL },
-  { cft_strg, "root",		NULL,		NULL,NULL },
-  { cft_strg, "vga",		NULL,		NULL,NULL },
-  { cft_flag, "lock",		NULL,		NULL,NULL },
-  { cft_strg, "ramdisk",	NULL,		NULL,NULL },
-  { cft_strg, "password",	NULL,		NULL,NULL },
-  { cft_flag, "restricted",	NULL,		NULL,NULL },
-  { cft_flag, "mandatory",	NULL,		NULL,NULL },
-  { cft_flag, "optional",	NULL,		NULL,NULL },
-  { cft_flag, "read-only",	NULL,		NULL,NULL },
-  { cft_flag, "read-write",	NULL,		NULL,NULL },
-  { cft_strg, "append",		NULL,		NULL,NULL },
-  { cft_flag, "nowarn",		NULL,		NULL,NULL },
+  { cft_strg, "disktab",	NULL,		NULL,NULL },
+  { cft_flag, "el-torito-bootable-cd",	NULL,		NULL,NULL },
   { cft_strg, "fallback",	NULL,		NULL,NULL },
+  { cft_flag, "fix-table",     	NULL,		NULL,NULL },
+  { cft_strg, "force-backup",	NULL,		NULL,NULL },
+  { cft_flag, "geometric",      NULL,           NULL,NULL }, 
+  { cft_flag, "ignore-table",   NULL,		NULL,NULL },
   { cft_strg, "initrd",		NULL,		NULL,NULL },
-  { cft_flag, "change-rules",	do_cr,		NULL,NULL },
-  { cft_flag, "single-key",	NULL,		NULL,NULL },
+  { cft_strg, "install",	NULL,		NULL,NULL },
   { cft_strg, "keytable",	NULL,		NULL,NULL },
+  { cft_flag, "large-memory",	NULL,		NULL,NULL },
+  { cft_flag, "lba32",          NULL,           NULL,NULL }, 
+  { cft_flag, "linear",     	NULL,		NULL,NULL },
   { cft_strg, "loader",		NULL,		NULL,NULL },
+  { cft_flag, "lock",		NULL,		NULL,NULL },
+  { cft_flag, "mandatory",	NULL,		NULL,NULL },
+  { cft_strg, "map",		NULL,		NULL,NULL },
+  { cft_flag, "master-boot",	NULL,		NULL,NULL },
   { cft_strg, "menu-scheme",	NULL,		NULL,NULL },
   { cft_strg, "menu-title",	NULL,		NULL,NULL },
-  { cft_strg, "bitmap",		NULL,		NULL,NULL },
-  { cft_strg, "bmp-table",	NULL,		NULL,NULL },
-  { cft_strg, "bmp-colors",	NULL,		NULL,NULL },
-  { cft_strg, "bmp-timer",	NULL,		NULL,NULL },
-  { cft_flag, "large-memory",	NULL,		NULL,NULL },
+  { cft_strg, "message",	NULL,		NULL,NULL },
+  { cft_flag, "nowarn",		NULL,		NULL,NULL },
+  { cft_flag, "optional",	NULL,		NULL,NULL },
+  { cft_strg, "password",	NULL,		NULL,NULL },
+  { cft_flag, "prompt",		NULL,		NULL,NULL },
+  { cft_strg, RAID_EXTRA_BOOT,	NULL,		NULL,NULL },
+  { cft_strg, "ramdisk",	NULL,		NULL,NULL },
+  { cft_flag, "read-only",	NULL,		NULL,NULL },
+  { cft_flag, "read-write",	NULL,		NULL,NULL },
+  { cft_flag, "restricted",	NULL,		NULL,NULL },
+  { cft_strg, "root",		NULL,		NULL,NULL },
+  { cft_strg, "serial",		NULL,		NULL,NULL },
+  { cft_flag, "single-key",	NULL,		NULL,NULL },
+  { cft_flag, "suppress-boot-time-BIOS-data",	NULL,	NULL,NULL },
+  { cft_strg, "timeout",	NULL,		NULL,NULL },
+  { cft_strg, "verbose",	NULL,		NULL,NULL },
+  { cft_strg, "vga",		NULL,		NULL,NULL },
+  { cft_strg, "vmdefault",	NULL,		NULL,NULL },
   { cft_end,  NULL,		NULL,		NULL,NULL }};
 
 CONFIG cf_all[] = {
-  { cft_strg, "label",		NULL,		NULL,NULL },
   { cft_strg, "alias",		NULL,		NULL,NULL },
+  { cft_flag, "bypass",		NULL,		NULL,NULL },
+  { cft_strg, "fallback",	NULL,		NULL,NULL },
+  { cft_strg, "label",		NULL,		NULL,NULL },
+  { cft_strg, "literal",	NULL,		NULL,NULL },
+  { cft_flag, "lock",		NULL,		NULL,NULL },
+  { cft_flag, "mandatory",	NULL,		NULL,NULL },
+  { cft_flag, "optional",	NULL,		NULL,NULL },
   { cft_strg, "password",	NULL,		NULL,NULL },
   { cft_flag, "restricted",	NULL,		NULL,NULL },
-  { cft_flag, "mandatory",	NULL,		NULL,NULL },
-  { cft_flag, "bypass",		NULL,		NULL,NULL },
-  { cft_flag, "optional",	NULL,		NULL,NULL },
-  { cft_flag, "lock",		NULL,		NULL,NULL },
-  { cft_strg, "fallback",	NULL,		NULL,NULL },
   { cft_flag, "single-key",	NULL,		NULL,NULL },
-  { cft_strg, "literal",	NULL,		NULL,NULL },
   { cft_flag, "vmdisable",	NULL,		NULL,NULL },
   { cft_flag, "vmwarn",		NULL,		NULL,NULL },
   { cft_end,  NULL,		NULL,		NULL,NULL }};
 
 CONFIG cf_kernel[] = {
-  { cft_strg, "root",		NULL,		NULL,NULL },
-  { cft_strg, "vga",		NULL,		NULL,NULL },
+  { cft_strg, "append",		NULL,		NULL,NULL },
+  { cft_strg, "initrd",		NULL,		NULL,NULL },
   { cft_strg, "ramdisk",	NULL,		NULL,NULL },
   { cft_flag, "read-only",	NULL,		NULL,NULL },
   { cft_flag, "read-write",	NULL,		NULL,NULL },
-  { cft_strg, "append",		NULL,		NULL,NULL },
-  { cft_strg, "initrd",		NULL,		NULL,NULL },
+  { cft_strg, "root",		NULL,		NULL,NULL },
+  { cft_strg, "vga",		NULL,		NULL,NULL },
   { cft_link, NULL,		&cf_all,	NULL,NULL }};
 
 CONFIG cf_image[] = {
@@ -139,20 +143,21 @@ CONFIG cf_image[] = {
   { cft_link, NULL,		&cf_kernel,	NULL,NULL }};
 
 CONFIG cf_other[] = {
+  { cft_strg, "boot-as",	NULL,		NULL,NULL },
+  { cft_flag, "change",		do_change,	NULL,NULL },
   { cft_strg, "loader",		NULL,		NULL,NULL },
+  { cft_strg, "map-drive",	do_map_drive,	NULL,NULL },
+  { cft_flag, "master-boot",	NULL,		NULL,NULL },
   { cft_strg, "table",		NULL,		NULL,NULL },
   { cft_flag, "unsafe",		NULL,		NULL,NULL },
-  { cft_flag, "master-boot",	NULL,		NULL,NULL },
-  { cft_strg, "map-drive",	do_map_drive,	NULL,NULL },
-  { cft_flag, "change",		do_change,	NULL,NULL },
   { cft_link, NULL,		&cf_all,	NULL,NULL }};
 
 CONFIG cf_disk[] = {
-  { cft_flag, "inaccessible",	NULL,		NULL,NULL },
   { cft_strg, "bios",		NULL,		NULL,NULL },
-  { cft_strg, "sectors",	NULL,		NULL,NULL },
-  { cft_strg, "heads",		NULL,		NULL,NULL },
   { cft_strg, "cylinders",	NULL,		NULL,NULL },
+  { cft_strg, "heads",		NULL,		NULL,NULL },
+  { cft_flag, "inaccessible",	NULL,		NULL,NULL },
+  { cft_strg, "sectors",	NULL,		NULL,NULL },
   { cft_end,  NULL,		NULL,		NULL,NULL }};
 
 CONFIG cf_partitions[] = {
@@ -168,40 +173,40 @@ CONFIG cf_map_drive[] = {
   { cft_end,  NULL,		NULL,		NULL,NULL }};
 
 CONFIG cf_change_rules[] = {
-  { cft_strg, "type",		do_cr_type,	NULL,NULL },
   { cft_flag, "reset",		do_cr_reset,	NULL,NULL },
+  { cft_strg, "type",		do_cr_type,	NULL,NULL },
   { cft_end,  NULL,		NULL,		NULL,NULL }};
 
 CONFIG cf_change_rule[] = {
-  { cft_strg, "normal",		NULL,		NULL,NULL },
   { cft_strg, "hidden",		NULL,		NULL,NULL },
+  { cft_strg, "normal",		NULL,		NULL,NULL },
   { cft_end,  NULL,		NULL,		NULL,NULL }};
 
 CONFIG cf_change[] = {
-  { cft_strg, "partition",	do_cr_part,	NULL,NULL },
   { cft_flag, "automatic",	do_cr_auto,	NULL,NULL },
+  { cft_strg, "partition",	do_cr_part,	NULL,NULL },
   { cft_end,  NULL,		NULL,		NULL,NULL }};
 
 CONFIG cf_change_dsc[] = {
-  { cft_strg, "set",		NULL,		NULL,NULL },
   { cft_flag, "activate",	NULL,		NULL,NULL },
   { cft_flag, "deactivate",	NULL,		NULL,NULL },
+  { cft_strg, "set",		NULL,		NULL,NULL },
   { cft_end,  NULL,		NULL,		NULL,NULL }};
 
 CONFIG cf_bitmap[] = {
   { cft_strg, "bitmap",		NULL,		NULL,NULL },
-  { cft_strg, "bmp-table",	NULL,		NULL,NULL },
   { cft_strg, "bmp-colors",	NULL,		NULL,NULL },
+  { cft_strg, "bmp-table",	NULL,		NULL,NULL },
   { cft_strg, "bmp-timer",	NULL,		NULL,NULL },
   { cft_end,  NULL,		NULL,		NULL,NULL }};
 
 
 #if NEW_PARSE
 
-static CONFIG *keywords[] = {cf_top, /* cf_identify, */ cf_options, cf_all,
+static CONFIG *keywords[] = {cf_top, cf_identify, cf_options, cf_all,
 	cf_kernel, cf_image, cf_other, cf_disk, cf_partitions, cf_partition,
 	cf_map_drive, cf_change_rules, cf_change_rule, cf_change, 
-	cf_change_dsc, NULL };
+	cf_change_dsc, cf_bitmap, NULL };
 
 #endif
 	
@@ -625,3 +630,20 @@ void cfg_bitmap_only(void)
     keywords[1] = NULL;
 }
 
+#if BETA_TEST
+void cfg_alpha_check(void)
+{
+    CONFIG **kw = keywords;
+    CONFIG *cfg;
+    
+    while ((cfg=*kw)) {
+	while (cfg[1].name) {
+	    if (strcmp(cfg[0].name, cfg[1].name) >= 0) {
+		die("cfg_alpha_check:  failure at '%s', '%s'", cfg[0].name, cfg[1].name);
+	    }
+	cfg++;
+	}
+    kw++;
+    }
+}
+#endif
