@@ -79,4 +79,10 @@ FILE *cfg_pw_open(void);
 void cfg_bitmap_only(void);
 /* allow only the "bitmap" keywords */
 
+#if BETA_TEST
+void cfg_alpha_check(void);
+/* check for tables in alphabetical order */
+#endif
+
+
 #endif
