@@ -19,21 +19,19 @@ source directory.
 #include "lilo.h"
 #include "common.h"
 
-
+#ifndef SHS_MAIN
 EXTRA extra;
 int verbose = 0,test = 0,compact = 0,linear = 0, raid_flags = 0, zflag = 0,
-      ireloc = 0, force_fs = 0,
+      ireloc = 0, force_fs = 0, extended_pt = 0,
       nowarn = 0, lba32 = 0, autoauto = 0, passw = 0, geometric = 0, eflag = 0;
+DL_BIOS bios_passes_dl = DL_NOT_SET;
 
 int bios_boot, bios_map;
 unsigned short drv_map[DRVMAP_SIZE+1]; /* fixup maps ... */
 int curr_drv_map;
 unsigned long prt_map[PRTMAP_SIZE+1];
 int curr_prt_map;
-#if 0
-unsigned long crc_polynomial[MAX_PW_CRC] = { 
-	CRC_POLY1, CRC_POLY2, CRC_POLY3, CRC_POLY4 };
-#endif	
+
 
 /*volatile*/ void pdie(char *msg)
 {
@@ -137,15 +135,19 @@ int stat_equal(struct stat *a,struct stat *b)
     return a->st_dev == b->st_dev && a->st_ino == b->st_ino;
 }
 
-/* calculate a CRC-32 polynomial */
+#endif	/*  !SHS_MAIN */
 
-unsigned long crc32 (unsigned char *cp, int nsize, unsigned long polynomial)
+
+/* accumulate a partial CRC-32 */
+
+unsigned long crc32partial(unsigned char *cp, int nsize,
+			unsigned long polynomial, unsigned long *accum)
 {
    unsigned long poly, crc;
    int i;
    unsigned char ch;
 
-   crc = 0xFFFFFFFFUL;
+   crc = ~*accum;
    while (nsize--) {
       ch = *cp++;
       for (i=0; i<8; i++) {
@@ -154,7 +156,16 @@ unsigned long crc32 (unsigned char *cp, int nsize, unsigned long polynomial)
          crc = (crc<<1) ^ poly;
       }
    }
-   return ~crc;
+   return (*accum = ~crc);
+}
+
+
+/* calculate a CRC-32 polynomial */
+
+unsigned long crc32 (unsigned char *cp, int nsize, unsigned long polynomial)
+{
+    unsigned long crc = 0;
+    return crc32partial(cp, nsize, polynomial, &crc);
 }
 
 
@@ -172,3 +183,29 @@ void show_link(char *name)
     }
 }
 
+
+
+#ifdef SHS_MAIN
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
+
+int main(int argc, char *argv[])
+{
+    unsigned char buf[4096];
+    int fd, n;
+    unsigned long crc;
+    
+    fd = open(argv[1],O_RDONLY);
+    crc = 0;
+    n = read(fd,buf,sizeof(buf));
+    while (n>0) {
+	crc32partial(buf, n, CRC_POLY1, &crc);
+	n = read(fd,buf,sizeof(buf));
+    }
+    close(fd);
+
+    printf("0x%08x\n", (int)crc);
+    return 0;
+}
+#endif	/* SHS_MAIN */
