@@ -20,7 +20,6 @@ source directory.
 #include <asm/types.h>
 /*#include <linux/genhd.h>*/
 
-/*#include "lilo.h"*/
 
 #ifdef O_ACCMODE
 # define O_NOACCESS O_ACCMODE
@@ -124,39 +123,28 @@ source directory.
 							block	0
 ;*/    unsigned char cli;     /* clear interrupt flag instruction
 						par1_cli:	.blkb	1
-;*/    unsigned char jmp0, jmp1, nop;	/* short jump followed by a NOP
-						par1_jump:	.blkb	3
+;*/    unsigned char jmp0, jmp1;	/* short jump
+						par1_jump:	.blkb	2
+;*/    unsigned char stage;  /*
+						par1_stage:	.blkb	1
 ;*/    unsigned short code_length;  /* length of the first stage code
 						par1_code_len:	.blkb	2
 ;*/    char signature[4]; /* "LILO"
 						par1_signature:	.blkb	4
-;*/    unsigned short stage,version;  /*
-						par1_stage:	.blkb	2
+;*/    unsigned short version;  /*
 						par1_version:	.blkb	2
-;*/    unsigned char port; /* COM port. 0 = none, 1 = COM1, etc. !!! keep these two serial bytes together !!!
-;*/    unsigned char ser_param; /* RS-232 parameters, must be 0 if unused
-						par1_port:	.blkb	1	; referenced together
-						par1_ser_param:	.blkb	1	; **
+;*/    unsigned long map_stamp; /* timestamp for this installation (map creation)
+						par1_mapstamp:	.blkb	4
 ;*/    unsigned long raid_offset; /* raid partition/partition offset
 						par1_raid_offset: .blkb	4
 ;*/    unsigned long timestamp; /* timestamp for restoration
 						par1_timestamp:	.blkb	4
-;*/    unsigned long map_stamp; /* timestamp for this installation (map creation)
-						par1_mapstamp:	.blkb	4
 ;*/    unsigned long map_serial_no; /* volume serial no. / id containing the map file
 						par1_map_serial_no:	.blkb	4
-;*/    unsigned short timeout; /* 54 msec delay until input time-out, 0xffff: never 
-						par1_timeout:	.blkb	2
-;*/    unsigned short delay; /* delay: wait that many 54 msec units.
-						par1_delay:	.blkb	2
-;*/    SECTOR_ADDR keytab; /* keyboard translation table
-						par1_keytab:	.blkb	sa_size
-;*/    SECTOR_ADDR secondary; /* sectors of the second stage loader
-						par1_secondary:	.blkb	sa_size
-;*/    unsigned short msg_len; /* 0 if none
-						par1_msg_len:	.blkb	2
 ;*/    unsigned char prompt; /* FLAG_PROMPT=always, FLAG_RAID install
 						par1_prompt:	.blkb	1
+;*/    SECTOR_ADDR secondary; /* sectors of the second stage loader
+						par1_secondary:	.blkb	sa_size
 ;*/} BOOT_PARAMS_1; /* first stage boot loader 
 								.align	4
 						par1_size:
@@ -170,13 +158,26 @@ source directory.
 						par2_jump:	.blkb	6
 ;*/    char signature[4]; /* "LILO"
 						par2_signature:	.blkb	4
-;*/    unsigned short stage,version;	/*
-						par2_stage:	.blkb	2
+;*/    unsigned short version;	/*
 						par2_version:	.blkb	2
-;*/    short unused;			/*
-						par2_unused:	.blkb	2
 ;*/    unsigned long map_stamp;		/* time of creation of the map file
 						par2_mapstamp:	.blkb	4
+;*/    unsigned short stage;	/*
+						par2_stage:	.blkb	2
+;*/    unsigned char port; /* COM port. 0 = none, 1 = COM1, etc. !!! keep these two serial bytes together !!!
+;*/    unsigned char ser_param; /* RS-232 parameters, must be 0 if unused
+						par2_port:	.blkb	1	; referenced together
+						par2_ser_param:	.blkb	1	; **
+;*/    unsigned short timeout; /* 54 msec delay until input time-out, 0xffff: never 
+						par2_timeout:	.blkb	2
+;*/    unsigned short delay; /* delay: wait that many 54 msec units.
+						par2_delay:	.blkb	2
+;*/    unsigned short msg_len; /* 0 if none
+						par2_msg_len:	.blkb	2
+;*/    SECTOR_ADDR keytab; /* keyboard translation table
+						par2_keytab:	.blkb	sa_size
+;*/    unsigned char flag2;	/*	flags specific to the second stage loader
+						par2_flag2:	.blkb	1
 ;*/} BOOT_PARAMS_2; /* second stage boot loader
 								.align	4
 						par2_size:
@@ -184,6 +185,34 @@ source directory.
 
 
 
+
+;*/typedef struct {			/*
+							block	0
+;*/    char jump[6]; /* jump over the data
+						parC_jump:	.blkb	6
+;*/    char signature[4]; /* "LILO" 
+						parC_signature:	.blkb	4
+;*/    unsigned short stage,version; /* stage is 0x10
+						parC_stage:	.blkb	2
+						parC_version:	.blkb	2
+;*/    unsigned short offset; /* partition entry offset
+						parC_offset:	.blkb	2
+;*/    unsigned char drive; /* BIOS drive code
+						parC_drive:	.blkb	1
+;*/    unsigned char head; /* head; always 0
+						parC_head:	.blkb	1
+;*/    unsigned short drvmap; /* offset of drive map
+						parC_drvmap:	.blkb	2
+;*/    unsigned char ptable[PARTITION_ENTRY*PARTITION_ENTRIES]; /* part. table
+						parC_ptable:	.blkb	64
+;*/    unsigned short p_devmap[2];	/* pointer to device map filled in by second.S
+						parC_devmap:	.blkb	4
+;*/} BOOT_PARAMS_C; /* chain loader
+						parC_size:
+							endb
+
+
+
 ;*/typedef struct {					/*
 							block	0
 ;*/	char menu_sig[4];	/* "MENU" or "BMP4" signature, or NULs if not present
@@ -253,31 +282,22 @@ source directory.
 ;*/
 
 #ifndef LILO_ASM
-#if 0
-typedef struct {
-    char jump[6]; /* jump over the data */
-    char signature[4]; /* "LILO" */
-    unsigned short stage,version;
-} BOOT_PARAMS_2; /* second stage boot loader */
-#endif
 
 typedef struct {
-    char jump[6]; /* jump over the data */
-    char signature[4]; /* "LILO" */
-    unsigned short stage,version; /* stage is 0x10 */
-    unsigned short offset; /* partition entry offset */
-    unsigned char drive; /* BIOS drive code */
-    unsigned char head; /* head; always 0 */
-    unsigned short drvmap; /* offset of drive map */
-    unsigned char ptable[PARTITION_ENTRY*PARTITION_ENTRIES]; /* part. table */
-} BOOT_PARAMS_C; /* chain loader */
-
-
+    unsigned char bootcode[MAX_BOOT_SIZE];
+    unsigned short mbz;		/* must be zero */
+    long  volume_id;
+    unsigned short marker;	/* may be zero */
+    unsigned char part[PART_TABLE_SIZE];
+    unsigned short boot_ind;	/* 0xAA55 */
+} BOOT_VOLID;
+    
 typedef union {
     BOOT_PARAMS_1 par_1;
     BOOT_PARAMS_2 par_2;
     BOOT_PARAMS_C par_c;
     BOOT_PARAMS_DOS par_d;
+    BOOT_VOLID boot;
     unsigned char sector[SECTOR_SIZE];
 } BOOT_SECTOR;
 
@@ -297,6 +317,17 @@ typedef struct {
     unsigned char data[1];
 } BUILTIN_FILE;
 
+
+#ifdef LCF_FIRST6
+#pragma pack (2)
+typedef struct {
+    unsigned char device, flags;
+    unsigned long sector;
+} SECTOR_ADDR6;
+#pragma pack ()
+#endif
+
+
 #endif
 /*
 IMAGES_numerator = SECTOR_SIZE_asm*MAX_DESCR_SECS_asm - 4 - 1
@@ -324,24 +355,31 @@ typedef struct {
 
 #define alloc_t(t) ((t *) alloc(sizeof(t)))
 
-typedef enum {X_NULL=0, X_NONE, X_AUTO, X_MBR_ONLY, X_SPEC} EXTRA;
+typedef enum {X_NULL=0, X_NONE, X_AUTO, X_MBR_ONLY, X_MBR, X_SPEC} EXTRA;
 typedef enum {AD_ANY=0, AD_GEOMETRIC, AD_LINEAR, AD_LBA32=4} ADDR_MODE;
+typedef enum {DL_NOT_SET=0, DL_BAD=1, DL_UNKNOWN, DL_MAYBE, DL_GOOD} DL_BIOS;
 
 extern EXTRA extra;
 extern int verbose,test,compact,linear,nowarn,lba32,autoauto,passw,geometric;
 extern int bios_boot, bios_map;
-extern int ireloc, force_fs;
+extern int ireloc, force_fs, extended_pt;
+extern DL_BIOS bios_passes_dl;
 extern int boot_dev_nr,raid_flags,do_md_install,ndisk,zflag,eflag;
 extern unsigned short drv_map[DRVMAP_SIZE+1]; /* needed for fixup maps */
 extern int curr_drv_map;
 extern unsigned long prt_map[PRTMAP_SIZE+1];
 extern int curr_prt_map, config_read;
 extern unsigned long serial_no[MAX_BIOS_DEVICES];
-#if 0
-extern unsigned long crc_polynomial[MAX_PW_CRC];
-#endif
 extern char *config_file;
 extern FILE *errstd;
+extern char *identify;	/* in identify.c */
+
+#define crc(a,b) (~crc32((a),(b),CRC_POLY1))
+#define brev(x)  \
+ { register unsigned short u1,u2; register unsigned long u=(x); u1=u; u2=u>>16; \
+ u1=(u1>>8)|(u1<<8); u2=(u2>>8)|(u2<<8); x=(unsigned long)u1<<16|u2; }
+#define cc(x) crc(x.data,x.size)
+#define comma (cc(First)|cc(Second)|cc(Third)|cc(Bitmap)|cc(Chain)|cc(Mbr)|cc(Mbr2))
 
 
 /*volatile*/ void pdie(char *msg);
@@ -377,6 +415,13 @@ int stat_equal(struct stat *a, struct stat *b);
    the same file, zero if they don't. */
 
 
+
+unsigned long crc32partial(unsigned char *cp, int nsize,
+			unsigned long polynomial, unsigned long *accum);
+/* accumulate a partial CRC-32 */
+
+
+
 unsigned long crc32 (unsigned char *cp, int nsize, unsigned long polynomial);
 /* calculate a CRC-32 polynomial */
 
