@@ -30,6 +30,7 @@ source directory.
 #include "lilo.h"
 #include "common.h"
 #include "device.h"
+#include "raid.h"
 #include "geometry.h"
 #include "cfg.h"
 #include "md-int.h"
@@ -39,12 +40,24 @@ source directory.
 #ifndef REISERFS_SUPER_MAGIC
 #define REISERFS_SUPER_MAGIC 0x52654973
 #endif
+
+#if 0
 #ifndef REISERFS_SUPER_MAGIC_STRING
 #define REISERFS_SUPER_MAGIC_STRING "ReIsErFs"
 #endif
+#endif
+
 #ifndef REISERFS_IOC_UNPACK
 #define REISERFS_IOC_UNPACK		_IOW(0xCD,1,long)
 #endif
+
+#ifndef REISER4_SUPER_MAGIC
+#define REISER4_SUPER_MAGIC  0x52345362
+ /* (*(__u32 *)"R4Sb"); */
+#endif
+#ifndef REISER4_IOC_UNPACK
+#define REISER4_IOC_UNPACK              _IOW(0xCD,1,long)
+#endif
 #endif
 
 #ifdef LCF_LVM
@@ -161,8 +174,8 @@ void do_partition(void)
     cfg_init(cf_partitions);
 }
 
-
-int has_partitions(dev_t dev)
+#if BETA_TEST
+int has_partitions_beta(dev_t dev)
 {
     int major = MAJOR(dev);
     
@@ -190,6 +203,53 @@ int has_partitions(dev_t dev)
 
     return 0;
 }
+#endif
+
+unsigned char max_partno[] = {
+/*
+  0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
+*/
+  0,  0,  0, 63,  0,  0,  0,  0, 15,  0,  0,  0,  0, 63,  0,  0,   /*  0x   */
+  0,  0,  0,  0,  0, 63, 63,  0,  0,  0,  0,  0,  0,  0,  0,  0,
+  0, 63, 63,  0, 63,  0,  0,  0,  0,  0,  0,  0, 15,  0,  0,  0,
+  7,  7,  7,  7,  7,  7,  7,  7, 63, 63,  0,  0, 15, 15, 15, 15,
+
+  0, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,   /*  4x   */
+ 15, 15, 15, 15, 15, 15, 15, 15, 63, 63, 63, 63,  0, 15,  0,  0,
+  0,  0,  0,  0,  0, 15,  0,  0, 15, 15, 15, 15, 15, 15, 15, 15,
+  7,  0, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
+/*
+  0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
+*/
+#if 0
+  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   /*  8x   */
+  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
+  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
+  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
+
+  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   /*  Cx   */
+  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
+  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
+  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0
+/*
+  0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
+*/
+#endif
+};
+
+int has_partitions(dev_t dev)
+{
+   int major = MAJOR(dev);
+   int ret;
+   
+   if (
+   	major < sizeof(max_partno) &&
+   	(ret=max_partno[major])
+   					)  ret ^= 0xFFFF;
+   else ret = 0;
+   
+   return ret;
+}
 
 
 void do_disk(void)
@@ -201,9 +261,9 @@ void do_disk(void)
     entry = alloc_t(DT_ENTRY);
     disk = cfg_get_strg(cf_options,"disk");
     if (stat(disk,&st) < 0) die("stat %s: %s",disk,strerror(errno));
-    if (!S_ISBLK(st.st_mode) || ((MINOR(st.st_rdev) & P_MASK(st.st_rdev)) &&
-      has_partitions(st.st_rdev)))
-	die("RSN: %s is not a whole disk device",disk);
+    if (!S_ISBLK(st.st_mode) || 
+    	(has_partitions(st.st_rdev) && (MINOR(st.st_rdev) & P_MASK(st.st_rdev))))
+		die(" '%s' is not a whole disk device",disk);
     entry->device = st.st_rdev;
     cfg_init(cf_disk);
     (void) cfg_parse(cf_disk);
@@ -231,11 +291,17 @@ void do_disk(void)
     }
     entry->cylinders = cylinders ? to_number(cylinders) : -1;
     entry->start = 0;
-    for (walk = disktab; walk; walk = walk->next)
+    for (walk = disktab; walk; walk = walk->next) {
 	if (entry->device == walk->device)
-	    die("Duplicate geometry definition for %s",disk);
+	    die("Duplicate \"disk =\" definition for %s",disk);
+    }
     entry->next = disktab;
     disktab = entry;
+    if (verbose >= 6) {
+	printf("do_disk: %s %04X 0x%02X  %d:%d:%d\n",
+		disk, entry->device, entry->bios, entry->cylinders,
+		entry->heads, entry->sectors);
+    }
     cfg_init(cf_partitions);
     (void) cfg_parse(cf_partitions);
     cfg_unset(cf_options,"disk");
@@ -309,7 +375,7 @@ static int last_dev(int major,int increment)
     DEVICE dev;
     int minor;
 
-    for (minor = 0; dev_open(&dev,(major << 8) | minor,-1); minor += increment)
+    for (minor = 0; dev_open(&dev,(major << 8) | minor,O_BYPASS); minor += increment)
 	if (exists(dev.name)) dev_close(&dev);
         else {
 	    dev_close(&dev);
@@ -320,8 +386,14 @@ static int last_dev(int major,int increment)
     DEVICE dev;
     int devs;
 
+#if 0
     for (devs = 0; devs < 2 && dev_open(&dev,(major << 8) | (increment*devs),
-      -1); devs++)
+      O_BYPASS); devs++)
+#else
+    for (devs = 0;
+	devs < 2 && dev_open(&dev,MKDEV(major,increment*devs),O_BYPASS);
+	    devs++)
+#endif
 	if (exists(dev.name)) dev_close(&dev);
         else {
 	    dev_close(&dev);
@@ -413,29 +485,39 @@ void evms_bmap(struct evms_get_bmap_t *ebm)
 #endif
 
 
-static void geo_query_dev(GEOMETRY *geo,int device,int all)
+void geo_query_dev(GEOMETRY *geo,int device,int all)
 {
     DEVICE dev;
     int fd,get_all;
     struct floppy_struct fdprm;
     struct hd_geometry hdprm;
 
-    get_all = all || MAJOR(device) != MAJOR_FD;
+    if (verbose>=5) printf("geo_query_dev: device=%04X\n", device);
+#if 0
+/*  Werner's original */
+    get_all = all || MAJOR(device) != MAJOR_FD; */
+#else
+/* simplify the condition -- JRC 2003-06-04 */
+    get_all = all;
+#endif
     if (!MAJOR(device))
 	die("Trying to map files from unnamed device 0x%04x (NFS ?)",device);
     if (device == MAJMIN_RAM)
 	die("Trying to map files from your RAM disk. "
 	  "Please check -r option or ROOT environment variable.");
-    if (get_all) fd = dev_open(&dev,device,O_NOACCESS);
-    else fd = -1; /* pacify GCC */
+    if (get_all) {
+	fd = dev_open(&dev,device,O_NOACCESS);
+    }
+    else {
+	fd = -1; /* pacify GCC */
+	geo->heads = geo->cylinders = geo->sectors = 1;
+	geo->start = 0;
+	geo->device = -1;
+    }
     switch (MAJOR(device)) {
 	case MAJOR_FD:
 	    geo->device = device & 3;
-	    if (!get_all) {
-		geo->heads = geo->cylinders = geo->sectors = 1;
-		geo->start = 0;
-		break;
-	    }
+	    if (!get_all) break;
 	    if (ioctl(fd,FDGETPRM,&fdprm) < 0)
 		die("geo_query_dev FDGETPRM (dev 0x%04x): %s",device,
 		  strerror(errno));
@@ -457,6 +539,9 @@ static void geo_query_dev(GEOMETRY *geo,int device,int all)
 	case MAJOR_ESDI:
 	case MAJOR_XT:
 	case MAJOR_ACORN:
+	    geo->device = 0x80 + (MINOR(device) >> 6) +
+		    (MAJOR(device) == MAJOR_HD ? 0 : last_dev(MAJOR_HD,64));
+	    if (!get_all) break;
 	    if (ioctl(fd,HDIO_GETGEO,&hdprm) < 0)
 		die("geo_query_dev HDIO_GETGEO (dev 0x%04x): %s",device,
 		  strerror(errno));
@@ -464,9 +549,6 @@ static void geo_query_dev(GEOMETRY *geo,int device,int all)
 	    geo->cylinders = hdprm.cylinders;
 	    geo->sectors = hdprm.sectors;
 	    geo->start = hdprm.start;
-	    if ((geo->device = bios_device(geo, device)) < 0)
-		geo->device = 0x80 + (MINOR(device) >> 6) +
-		    (MAJOR(device) == MAJOR_HD ? 0 : last_dev(MAJOR_HD,64));
 	    break;
 	case MAJOR_SD:
 	case MAJOR_SD2:
@@ -476,6 +558,8 @@ static void geo_query_dev(GEOMETRY *geo,int device,int all)
 	case MAJOR_SD6:
 	case MAJOR_SD7:
 	case MAJOR_SD8:
+	    geo->device = 0x80 + last_dev(MAJOR_HD,64) + (MINOR(device) >> 4);
+	    if (!get_all) break;
 	    if (ioctl(fd,HDIO_GETGEO,&hdprm) < 0)
 		die("geo_query_dev HDIO_GETGEO (dev 0x%04x): %s",device,
 		  strerror(errno));
@@ -486,8 +570,6 @@ static void geo_query_dev(GEOMETRY *geo,int device,int all)
 	    geo->cylinders = hdprm.cylinders;
 	    geo->sectors = hdprm.sectors;
 	    geo->start = hdprm.start;
-	    if ((geo->device = bios_device(geo, device)) < 0)
-		geo->device = 0x80 + last_dev(MAJOR_HD,64) + (MINOR(device) >> 4);
 	    break;
 	case MAJOR_DAC960:
 	case MAJOR_DAC960+1:
@@ -498,6 +580,8 @@ static void geo_query_dev(GEOMETRY *geo,int device,int all)
 	case MAJOR_DAC960+6:
 	case MAJOR_DAC960+7:
 	case MAJOR_IBM_iSER:
+	    geo->device = 0x80 + last_dev(MAJOR_HD,64) + (MINOR(device) >> 3);
+	    if (!get_all) break;
 	    if (ioctl(fd,HDIO_GETGEO,&hdprm) < 0)
 		die("geo_query_dev HDIO_GETGEO (dev 0x%04x): %s",device,
 		  strerror(errno));
@@ -508,8 +592,6 @@ static void geo_query_dev(GEOMETRY *geo,int device,int all)
 	    geo->cylinders = hdprm.cylinders;
 	    geo->sectors = hdprm.sectors;
 	    geo->start = hdprm.start;
-	    if ((geo->device = bios_device(geo, device)) < 0)
-		geo->device = 0x80 + last_dev(MAJOR_HD,64) + (MINOR(device) >> 3);
 	    break;
 	case MAJOR_AMI_HYP:
 	case MAJOR_HPT370:
@@ -543,6 +625,8 @@ static void geo_query_dev(GEOMETRY *geo,int device,int all)
 	case MAJOR_I2O+5:
 	case MAJOR_I2O+6:
 	case MAJOR_I2O+7:
+	    geo->device = 0x80 + last_dev(MAJOR_HD,64) + (MINOR(device) >> 4);
+	    if (!get_all) break;
 	    if (ioctl(fd,HDIO_GETGEO,&hdprm) < 0)
 		die("geo_query_dev HDIO_GETGEO (dev 0x%04x): %s",device,
 		  strerror(errno));
@@ -553,14 +637,13 @@ static void geo_query_dev(GEOMETRY *geo,int device,int all)
 	    geo->cylinders = hdprm.cylinders;
 	    geo->sectors = hdprm.sectors;
 	    geo->start = hdprm.start;
-	    if ((geo->device = bios_device(geo, device)) < 0)
-		geo->device = 0x80 + last_dev(MAJOR_HD,64) + (MINOR(device) >> 4);
 	    break;
 
 	default:
 	    die("Sorry, don't know how to handle device 0x%04x",device);
     }
     if (get_all) dev_close(&dev);
+    if (verbose>=5) printf("exit geo_query_dev\n");
 }
 
 
@@ -606,7 +689,10 @@ int is_first(int device)
 	case MAJOR_SD8:
 	case MAJOR_AMI_HYP:
 	case MAJOR_HPT370:
-	case MAJOR_FL:
+	case MAJOR_EXPR+0:
+	case MAJOR_EXPR+1:
+	case MAJOR_EXPR+2:
+	case MAJOR_EXPR+3:
 	case MAJOR_NFTL:
 	case MAJOR_SMART2+0:
 	case MAJOR_SMART2+1:
@@ -656,6 +742,7 @@ void geo_get(GEOMETRY *geo,int device,int user_device,int all)
     DT_ENTRY *walk;
     int inherited,keep_cyls,is_raid=0;
 
+    if (verbose>=5) printf("geo_get: device %04X, all=%d\n", device, all);
 #ifdef LCF_LVM
     /*
      * Find underlying device (PV) for LVM.  It is OK if the underlying PV is
@@ -725,31 +812,76 @@ void geo_get(GEOMETRY *geo,int device,int user_device,int all)
    	for (pass = 0; pass < raid_limit; pass++) {
 	    md_disk_info.number = pass;
 	    if (ioctl(md_fd,GET_DISK_INFO,&md_disk_info) < 0)
+#if BETA_TEST
+	    {
+		printf("(raid) GET_DISK_INFO: failed for pass=%d\n", pass);
+		continue;
+	    }
+#else
 	        die("GET_DISK_INFO: %s", mdxxx);
+#endif
 	    if (!(md_disk_info.state & (1 << MD_DISK_FAULTY))) {
+#if 1
 		is_raid = (device==boot_dev_nr);
-	        device = (md_disk_info.major << 8) | md_disk_info.minor;
+#else
+/* this change may be in error; the correct comparison is == */
+		is_raid = (device!=boot_dev_nr);
+#endif
+	        device = MKDEV(md_disk_info.major, md_disk_info.minor);
 		break;
 	    }
 	}
 	close(md_fd);
     }
+
+#if BETA_TEST
+	if (verbose>=5) printf("geo_get(2):  device=%04X, all=%d\n", device, all);
+#endif
+
+/* if using hard disk, scan the devices in /proc/partitions */
+    if (has_partitions(device) && all)   pf_hard_disk_scan();
+
+
+
+
     for (walk = disktab; walk; walk = walk->next)
 	if (walk->device == device) break;
     inherited = !walk && !old_disktab;
+#if BETA_TEST
+    if (verbose>=5) printf("inherited=%d\n", inherited);
+#endif
     if (inherited)
 	for (walk = disktab; walk; walk = walk->next)
 	    if (walk->device == (device & D_MASK(device))) break;
+#if BETA_TEST
+    if (verbose>=5) printf("walk=%08x\n", (int)walk);
+#endif
 
+#if 1
 /* add 'all' to conditional below -- JRC 2002-08-20 */
     if (walk && !walk->heads && all)
+#else
+/* Werner's original conditional */
+    if (walk && !walk->heads)
+#endif
 	die("Device 0x%04X: Configured as inaccessible.\n",device);
     keep_cyls = !walk || walk->bios == -1 || walk->heads == -1 ||
       walk->sectors == -1 || inherited || walk->start == -1;
+#if BETA_TEST
+    if (verbose>=5) printf("inherited=%d  keep_cyls=%d\n", inherited, keep_cyls);
+#endif
 
+#if 1
 /* add 'all' to conditional below -- JRC 2002-08-20 */
     if (keep_cyls && (all || MAJOR(device)==MAJOR_FD) ) {
+#else
+/* Werner's original conditional */
+    if (keep_cyls) {
+#endif
 	geo_query_dev(geo,device,all);
+	
+	if (all) bios_device(geo, device);
+	
 	if (!nowarn && (geo->device & 0x7f) >= bios_max_devs() &&
 	  user_device == -1 && (!walk || walk->bios == -1))
 	    fprintf(errstd,"Warning: BIOS drive 0x%02x may not be accessible\n",
@@ -789,7 +921,7 @@ void geo_get(GEOMETRY *geo,int device,int user_device,int all)
         "   Use of the 'lba32' option may help on newer (EDD BIOS) systems.\n",
 	  device,BIOS_MAX_CYLS);
     }
-    if (verbose > 2) {
+    if (verbose >= 3) {
 	printf("Device 0x%04x: BIOS drive 0x%02x, %d heads, %d cylinders,\n",
 	  device,geo->device,geo->heads,geo->cylinders == -1 ? BIOS_MAX_CYLS :
 	  geo->cylinders);
@@ -800,7 +932,10 @@ void geo_get(GEOMETRY *geo,int device,int user_device,int all)
 
 /* make the serial number association */
     if (!is_raid) register_bios(geo->device, device);
-}
+    else geo->device = md_bios;		/* 22.5.7 add this else */
+    
+    return;
+} /* end of geo_get */
 
 
 int geo_open(GEOMETRY *geo,char *name,int flags)
@@ -822,7 +957,7 @@ int geo_open(GEOMETRY *geo,char *name,int flags)
     if (!S_ISREG(st.st_mode) && !S_ISBLK(st.st_mode))
 	die("%s: neither a reg. file nor a block dev.",name);
     geo->dev = S_ISREG(st.st_mode) ? st.st_dev : st.st_rdev;
-#if VERSION_MINOR>=50
+#if BETA_TEST
 	if (verbose>=4) {
 		printf("geo_open: (%s) st_dev(file)=%04X  st_rdev(blk)=%04X\n",
 		name,
@@ -830,6 +965,7 @@ int geo_open(GEOMETRY *geo,char *name,int flags)
 		(int)st.st_rdev );
 	}
 #endif
+
     geo_get(geo, geo->dev, user_dev, 1);
     geo->file = S_ISREG(st.st_mode) ? st.st_dev : 0;
     geo->boot = 0;
@@ -857,6 +993,7 @@ int geo_open_boot(GEOMETRY *geo,char *name)
 {
     struct stat st;
 
+    if (verbose>=5) printf("geo_open_boot: %s\n", name);
     if (stat(name,&st) < 0) die("stat %s: %s",name,strerror(errno));
     if (!S_ISREG(st.st_mode) && !S_ISBLK(st.st_mode))
 	die("%s: neither a reg. file nor a block dev.",name);
@@ -915,6 +1052,24 @@ int geo_comp_addr(GEOMETRY *geo,int offset,SECTOR_ADDR *addr)
 			die("Cannot unpack ReiserFS file");
 		if (verbose > 3) printf("fd %d: REISERFS_IOC_UNPACK\n", geo->fd);
 	    }
+        /* Forcing reiser4 to perform tail2extent converstion */
+           if (buf.f_type == REISER4_SUPER_MAGIC) {
+               if (ioctl (geo->fd, REISER4_IOC_UNPACK, 1) != 0)
+                       die("Cannot unpack Reiser4 file");
+               if (verbose > 3) printf("fd %d: REISER4_IOC_UNPACK\n", geo->fd);
+
+           /* 
+               As we may have the situation when extent will be included
+               into transaction, and its item(s) will not be have the real block
+               numbers assigned, we should perform fsync() in order to guarantee,
+               that current atom is flushed and real block numbers assigned to 
+               the extent(s) file was converted in.
+           */
+		if (fdatasync(geo->fd) != 0)
+		    die("Cannot perform fdatasync");
+           
+		if (verbose > 3) printf("fd %d: fdatasync()\n", geo->fd);
+           }
 #endif
 	if (ioctl(geo->fd,FIBMAP,&block) < 0) pdie("ioctl FIBMAP");
 	if (!block) {
@@ -1025,8 +1180,79 @@ int geo_find(GEOMETRY *geo,SECTOR_ADDR addr)
 	if (geo_comp_addr(geo,i*SECTOR_SIZE,&here))
 	    if (here.sector == addr.sector && here.track == addr.track &&
 	      here.device == addr.device && here.head == addr.head) {
-		if (lseek(geo->fd,i*SECTOR_SIZE,0) < 0) return 0;
+		if (lseek(geo->fd,i*SECTOR_SIZE,SEEK_SET) < 0) return 0;
 		else return 1;
 	    }
     return 1;
 }
+
+
+#if 0
+int geo_devscan(int device)
+{
+    DT_ENTRY *walk;
+    unsigned int mask, codes = 0;
+    int bios;
+    int maxbios = 0;
+    
+    device &= D_MASK(device);
+
+/* mark those BIOS codes that are already used */
+    for (walk=disktab; walk; walk=walk->next) {
+	if (has_partitions(walk->device) && walk->bios != -1) {
+	    bios = walk->bios & 0x7F;
+	    if (bios >= 4*sizeof(codes) ) die("BIOS code %02X is too big (device %04X)", bios, device);
+	    codes |= 1 << bios;	
+	}
+    }
+
+    bios = -1;
+/* extract BIOS code of master device, or -1 */
+    for (walk=disktab; walk; walk=walk->next) {
+	if (device == walk->device) {
+	    bios = walk->bios;
+	}
+    }
+    if (bios > maxbios) maxbios = bios;
+
+/* if device has no BIOS code assigned, assign the next one */
+    if (bios == -1)
+	for (bios=0x80, mask=1; mask; mask<<=1, bios++)
+		if (!(mask&codes)) break;
+    
+    if (bios > DEV_MASK) die("geo_devscan:  ran out of device codes");
+    
+    for (walk=disktab; walk; walk=walk->next) {
+	if (device == walk->device) {
+	    if (walk->bios == -1) walk->bios = bios;
+	    else bios = walk->bios;
+	    break;
+	}
+    }
+    if (bios > maxbios) maxbios = bios;
+    
+    if (verbose >= 2) printf("geo_devscan:  maxbios = %02X\n", maxbios);
+    
+    if (walk) return maxbios;	/* there was an entry in the disktab */
+    
+    walk = alloc_t(DT_ENTRY);
+    walk->device = device;
+    walk->bios = bios;
+    walk->sectors = walk->heads = walk->cylinders = walk->start = -1;
+    walk->next = disktab;
+    disktab = walk;
+    if (verbose>=2)
+	printf("geo_devscan: arbitrary bios assignment  dev=%04X  bios=0x%02X\n",
+			device, bios);
+
+    for (walk=disktab; walk; walk=walk->next) {
+	if (device == (walk->device & D_MASK(walk->device))) {
+	    if (walk->bios != -1) walk->bios = bios;
+	}
+    }
+
+    return maxbios;
+}
+
+#endif
+
