@@ -20,15 +20,15 @@ source directory.
 #include <time.h>
 #include <limits.h>
 #include <sys/stat.h>
-#include <asm/page.h>
+/*#include <asm/page.h>*/
 
 #include "config.h"
 #include "lilo.h"
 #include "common.h"
 #include "raid.h"
 #include "boot.h"
-#include "temp.h"
 #include "device.h"
+#include "flags.i"
 #include "geometry.h"
 #include "map.h"
 #include "bsect.h"
@@ -36,10 +36,10 @@ source directory.
 #include "identify.h"
 #include "partition.h"
 #include "probe.h"
+#include "temp.h"
 #include "loader.h"
 #include "md-int.h"
 #include "edit.h"
-#include "flags.i"
 
 char *config_file;		/* actual name of the config file */
 int config_read;		/* readable by other than root */
@@ -70,7 +70,8 @@ static void show_other(int fd)
     prtmap = drvmap+2*(DRVMAP_SIZE+1);
     while (drvmap[0] && drvmap[1]) {
 	if (drvmap[0]==0xFF && drvmap[1]==0xFF) {
-	    printf("    Master-Boot:  BIOS drive will always boot as 0x80\n");
+	    if (drvmap[3]==0xFF) printf("    Master-Boot:  This BIOS drive will always appear as 0x80 (or 0x00)\n");
+	    else printf("    Boot-As:  This BIOS drive will always appear as 0x%02X\n", drvmap[3]);
 	    drvmap += 4;
 	} else {
 	    printf("    BIOS drive 0x%02X is mapped to 0x%02X\n",drvmap[0],
@@ -92,6 +93,7 @@ static void show_images(char *map_file)
     DESCR_SECTORS descrs;
     BOOT_SECTOR boot;
     MENUTABLE menu;
+    BOOT_PARAMS_2 param2;
     GEOMETRY geo;
     SECTOR_ADDR addr[4];
     char buffer[SECTOR_SIZE];
@@ -103,13 +105,17 @@ static void show_images(char *map_file)
 
     fd = geo_open(&geo,map_file,O_RDONLY);
     if (read(fd,buffer,SECTOR_SIZE) != SECTOR_SIZE)
-	die("read1 %s: %s",map_file,strerror(errno));
+	die("read cmdline %s: %s",map_file,strerror(errno));
     if (read(fd,(char*)&descrs,sizeof(descrs)) != sizeof(descrs))
-	die("read2 %s: %s",map_file,strerror(errno));
+	die("read descrs %s: %s",map_file,strerror(errno));
+    if (lseek(fd, SECTOR_SIZE, SEEK_CUR) <= 0)	/* skip zero sector */
+	die("lseek over zero sector %s: %s",map_file,strerror(errno));
+    if (read(fd,(char*)&param2,sizeof(param2)) != sizeof(param2))
+	die("read second params %s: %s",map_file,strerror(errno));
     if (lseek(fd, - sizeof(menu), SEEK_END) <= 0)
-	die("lseek %s: %s",map_file,strerror(errno));
+	die("lseek keytable %s: %s",map_file,strerror(errno));
     if (read(fd,(char*)&menu,sizeof(menu)) != sizeof(menu))
-	die("read3 %s: %s",map_file,strerror(errno));
+	die("read keytable %s: %s",map_file,strerror(errno));
     tlba32  = (descrs.d.descr[0].start.device & LBA32_FLAG) != 0;
     tlinear = !tlba32 && (descrs.d.descr[0].start.device & LINEAR_FLAG);
     if (tlinear != linear  ||  tlba32 != lba32) {
@@ -135,11 +141,11 @@ static void show_images(char *map_file)
 #endif
 	printf("Installed:  %s\n", ctime((time_t*)&boot.par_1.map_stamp));
 	printf("Global settings:\n");
-	tsecs = (boot.par_1.delay*2197+3999)/4000;
+	tsecs = (param2.delay*2197+3999)/4000;
 	printf("  Delay before booting: %d.%d seconds\n",tsecs/10,tsecs % 10);
-	if (boot.par_1.timeout == 0xffff) printf("  No command-line timeout\n");
+	if (param2.timeout == 0xffff) printf("  No command-line timeout\n");
 	else {
-	    tsecs = (boot.par_1.timeout*2197+3999)/4000;
+	    tsecs = (param2.timeout*2197+3999)/4000;
 	    printf("  Command-line timeout: %d.%d seconds\n",tsecs/10,
 	      tsecs % 10);
 	}
@@ -147,20 +153,22 @@ static void show_images(char *map_file)
 	else printf("  Enter boot prompt only on demand\n");
 	printf("  Boot-time BIOS data%s saved\n",
 		boot.par_1.prompt & FLAG_NOBD ? " NOT" : "");
+	printf("  Boot-time BIOS data auto-suppress write%s bypassed\n",
+		boot.par_1.prompt & FLAG_BD_OKAY ? "" : " NOT");
 	printf("  Large memory (>15M) is%s used to load initial ramdisk\n", 
 		boot.par_1.prompt & FLAG_LARGEMEM ? "" : " NOT");
 	printf("  %sRAID installation\n",
 		boot.par_1.prompt & FLAG_RAID ? "" : "Non-");
 	printf("  Boot device %s be used for the Map file\n",
 		boot.par_1.prompt & FLAG_MAP_ON_BOOT ? "WILL" : "will not");
-	if (!boot.par_1.port) printf("  Serial line access is disabled\n");
+	if (!param2.port) printf("  Serial line access is disabled\n");
 	else printf("  Boot prompt can be accessed from COM%d\n",
-	      boot.par_1.port);
-	if (!boot.par_1.msg_len) printf("  No message for boot prompt\n");
+	      param2.port);
+	if (!param2.msg_len) printf("  No message for boot prompt\n");
 	else if (!cfg_get_strg(cf_options,"bitmap"))
-	    printf("  Boot prompt message is %d bytes\n",boot.par_1.msg_len);
+	    printf("  Boot prompt message is %d bytes\n",param2.msg_len);
 	else printf("  Bitmap file is %d paragraphs (%d bytes)\n",
-			boot.par_1.msg_len, 16*boot.par_1.msg_len);
+			param2.msg_len, 16*param2.msg_len);
 	if (*(unsigned short *) buffer != DC_MAGIC || !buffer[2])
 	    printf("  No default boot command line\n");
 	else printf("  Default boot command line: \"%s\"\n",buffer+2);
@@ -171,7 +179,7 @@ static void show_images(char *map_file)
     }
     for (image = 0; image < MAX_IMAGES; image++)
 	if (*(name = descrs.d.descr[image].name)) {
-	    printf("%s%-" S(MAX_IMAGE_NAME) "s %s%s",verbose > 0 ? "  " : "",name,
+	    fprintf(stdout,"%s%-" S(MAX_IMAGE_NAME) "s %s%s",verbose > 0 ? "  " : "",name,
 	      image ? "" : "*",
 #ifdef LCF_VIRTUAL
 	      descrs.d.descr[image].flags & FLAG_VMDEFAULT ? "@" :
@@ -240,28 +248,15 @@ static void show_images(char *map_file)
 			}
 		    }
 #endif
-#if 0
-		    if (!descrs.d.descr[image].start_page)
-			printf("    Kernel is loaded \"low\"\n");
-		    else printf("    Kernel is loaded \"high\", at 0x%08lx\n",
-			  (unsigned long) descrs.d.descr[image].start_page*
-			  PAGE_SIZE);
-#else
 		    if (!(flags & FLAG_LOADHI))
 			printf("    Kernel is loaded \"low\"\n");
 		    else printf("    Kernel is loaded \"high\"\n");
-#endif
-#if 1
 		    if (!*(unsigned long *) descrs.d.descr[image].rd_size)
 			printf("    No initial RAM disk\n");
 		    else printf("    Initial RAM disk is %ld bytes\n",
 			  *(unsigned long *) descrs.d.descr[image].rd_size);
-#else
-		    if (!descrs.d.descr[image].rd_size)
-			printf("    No initial RAM disk\n");
-		    else printf("    Initial RAM disk is %d sectors\n",
-			  (int)descrs.d.descr[image].rd_size);
-#endif
+		    if (flags & FLAG_TOOBIG)
+			printf("       and is too big to fit between 4M-15M\n");
 		}
 		if (!geo_find(&geo,descrs.d.descr[image].start)) {
 		    printf("    Map sector not found\n");
@@ -331,14 +326,14 @@ static void usage(char *name)
     fprintf(errstd,"%12s[ -F ] [ -i boot_loader ] [ -m map_file ] [ -d delay ]\n","");
     fprintf(errstd,"%12s[ -v N | -v ... ] [ -t ] [ -s save_file | -S save_file ]\n",
       "");
-    fprintf(errstd,"%12s[ -p ][ -P fix | -P ignore ] [ -r root_dir ] [ -w ]\n","");
+    fprintf(errstd,"%12s[ -p ][ -P fix | -P ignore ] [ -r root_dir ] [ -w | -w+ ]\n","");
     fprintf(errstd,"%7s%s [ -C config_file ] [ -m map_file ] "
       "-R [ word ... ]\n","",name);
     fprintf(errstd,"%7s%s [ -C config_file ] -I name [ options ]\n","",name);
     fprintf(errstd,"%7s%s [ -C config_file ] [ -s save_file ] "
       "-u | -U [ boot_device ]\n","",name);
-    fprintf(errstd,"%7s%s -A /dev/XXX [ N ]\t\tactivate a partition\n","",name);
-    fprintf(errstd,"%7s%s -M /dev/XXX [ mbr_file ]\tinstall master boot record\n","",name);
+    fprintf(errstd,"%7s%s -A /dev/XXX [ N ]\t\tinquire/activate a partition\n","",name);
+    fprintf(errstd,"%7s%s -M /dev/XXX [ mbr | ext ]\tinstall master boot record\n","",name);
     fprintf(errstd,"%7s%s -T help \t\t\tlist additional options\n", "", name);
     fprintf(errstd,"%7s%s -X\t\t\t\tinternal compile-time options\n", "", name);
     fprintf(errstd,"%7s%s -V [ -v ]\t\t\tversion information\n\n","",name);
@@ -353,12 +348,13 @@ int main(int argc,char **argv)
 static char *bitmap_file;
     int query,more,version,uninstall,validate,activate,instmbr,geom;
     struct stat st;
-    int fd, temp;
+    int fd, temp, tell_early=0;
     long raid_offset;
 
     errstd = stderr;
 #if VERSION_MINOR>=50
     if (sizeof(MENUTABLE)!=256) die("MENUTABLE is not 256 bytes (common.h)");
+    cfg_alpha_check();
 #endif
     config_file = DFL_CONFIG;
     act1 = act2 = tell_param = 
@@ -368,12 +364,16 @@ static char *bitmap_file;
     verbose = -1;
     name = *argv;
     argc--;
+    
+    if (atexit( (void(*)(void)) sync)) die("atexit(sync)");
+    if (atexit( (void(*)(void)) purge)) die("atexit(purge)");
+    
     cfg_init(cf_options);
     while (argc && **++argv == '-') {
 	argc--;
       /* first those options with a mandatory parameter */
       /* Notably absent are "RuUvw" */
-	if (strchr("AbBCdDEfiImMPrsSTx", ch=(*argv)[1])) {
+	if (strchr("AbBCdDEfiImMPrsSTxZ", ch=(*argv)[1])) {
 	    if ((*argv)[2]) param = (*argv)+2;
 	    else {
 		param = *++argv;
@@ -381,6 +381,8 @@ static char *bitmap_file;
 	    }
 	} else { 
 	    param = NULL;
+	    if (strchr("cFglLpqtVXz", ch)	/* those with no args */
+	    	&& (*argv)[2]) usage(name);
 	}
 #if 0
 fprintf(errstd,"argc=%d, *argv=%s, ch=%c param=%s\n", argc, *argv, ch, param);
@@ -450,7 +452,7 @@ fprintf(errstd,"argc=%d, *argv=%s, ch=%c param=%s\n", argc, *argv, ch, param);
 	    case 'M':
 		instmbr = 1;
 		act1 = param;
-#ifndef LCF_BUILTIN
+#if !defined LCF_BUILTIN	|| 1
 		if (argc && argv[1][0] != '-') {
 		    act2 = *++argv;
 		    argc--;
@@ -461,11 +463,14 @@ fprintf(errstd,"argc=%d, *argv=%s, ch=%c param=%s\n", argc, *argv, ch, param);
 		passw = 1;	/* force re-gen of password file */
 		break;
 	    case 'P':
-		if (!strcmp(param,"fix"))
+		if (!strcasecmp(param,"fix"))
 		    cfg_set(cf_options,"fix-table",NULL,NULL);
-		else if (!strcmp(param,"ignore"))
+		else if (!strcasecmp(param,"ignore"))
 		    cfg_set(cf_options,"ignore-table",NULL,NULL);
-		else usage(name);
+		else if (!strcasecmp(param,"x"))
+		    extended_pt = 1;
+		else
+		    cfg_set(cf_options,param,NULL,NULL);
 		break;
 	    case 'q':
 		query = 1;
@@ -539,14 +544,7 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 		if ( (*argv)[2] == '+' ) nowarn = -1;
 		break;
 	    case 'x':
-		if (!strcmp(param,"none"))
-		    cfg_set(cf_options,RAID_EXTRA_BOOT,param,NULL);
-		else if (!strcmp(param,"auto"))
-		    cfg_set(cf_options,RAID_EXTRA_BOOT,param,NULL);
-		else if (!strcmp(param,"mbr-only"))
-		    cfg_set(cf_options,RAID_EXTRA_BOOT,param,NULL);
-		else
-		    cfg_set(cf_options,RAID_EXTRA_BOOT,param,NULL);
+		cfg_set(cf_options,RAID_EXTRA_BOOT,param,NULL);
 		break;
 	    case 'X':
 #if VERSION_MINOR>=50
@@ -581,16 +579,19 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 		sizeof(BOOT_PARAMS_DOS)
 		  );
 }
+#ifdef LCF_FIRST6
+	printf("\nSECTOR_ADDR6=%d\n", sizeof(SECTOR_ADDR6));
+#endif
+
 #endif
-#define crc(a,b) (~crc32((a),(b),CRC_POLY1))
-#define brev(x)  \
- { register unsigned short u1,u2; register unsigned long u=(x); u1=u; u2=u>>16; \
- u1=(u1>>8)|(u1<<8); u2=(u2>>8)|(u2<<8); x=(unsigned long)u1<<16|u2; }
-#define cc(x) crc(x.data,x.size)
-#define comma (cc(Second)|cc(Third)|cc(Bitmap))
 {
   unsigned long j;
   if (verbose>=5) {
+    if (crc(First.data,First.size)) {
+	j = crc(First.data,First.size-4);
+	brev(j);
+	printf("1=0x%lx\n", j);
+    }
     if (crc(Second.data,Second.size)) {
 	j = crc(Second.data,Second.size-4);
 	brev(j);
@@ -604,25 +605,48 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
     if (crc(Bitmap.data,Bitmap.size)) {
 	j = crc(Bitmap.data,Bitmap.size-4);
 	brev(j);
-	printf("B=0x%lx\n\n", j);
+	printf("B=0x%lx\n", j);
+    }
+    if (crc(Chain.data,Chain.size)) {
+	j = crc(Chain.data,Chain.size-4);
+	brev(j);
+	printf("C=0x%lx\n", j);
+    }
+    if (crc(Mbr.data,Mbr.size)) {
+	j = crc(Mbr.data,Mbr.size-4);
+	brev(j);
+	printf("M=0x%lx\n", j);
+    }
+    if (crc(Mbr2.data,Mbr2.size)) {
+	j = crc(Mbr2.data,Mbr2.size-4);
+	brev(j);
+	printf("N=0x%lx\n\n", j);
     }
   }
 }
 
+#if BETA_TEST
+{
+    int i, k=0;
+extern int has_partitions_beta(dev_t dev);	/* defined in geometry.c */
+
+
+    for (i=0; i<256; i++) {
+	k += ( has_partitions_beta(MKDEV(i,0)) != has_partitions(MKDEV(i,0)) );
+    }
+    printf("\nk = %d (should be 0)\n", k);
+}
+#endif
+
 #if VERSION_MINOR>=50
 {
 	dev_t dev; int i;
-	struct {
-		short a[0x1b6/2];
-		long  b;
-		short c;
-		char pt[64];
-		short d;
-		} x;
+	BOOT_VOLID x;
+	BOOT_SECTOR b;
 	dev = 0xFFF0;
 	i = dev;
 	printf("dev_t is %ssigned size=%d  i=%08X\n", i==0xFFF0 ? "un" : "", sizeof(dev_t), i);
-	printf("size of x = %d\n", sizeof(x));
+	printf("size of BOOT_VOLID = %d   BOOT_SECTOR = %d\n", sizeof(x), sizeof(b));
 	printf("Size of MENUTABLE = %d\n", sizeof(MENUTABLE));
 }
 #endif
@@ -632,6 +656,9 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 	    case 'z':
 		zflag++;	/* force zero of MBR 8-byte area */
 		break;
+	    case 'Z':
+		cfg_set(cf_options,"bios-passes-dl",param,NULL);
+		break;
 	    default:
 		usage(name);
 	}
@@ -657,11 +684,8 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 	printf(", Copyright (C) 1992-1998 Werner Almesberger\n"
 	       "Development beyond version 21 Copyright (C) 1999-2003 John Coffman\n"
 	       );
-        if (verbose>0) printf("Released %s%s and compiled at %s on %s.\n",
-	        VERSION_DATE, comma ? "," : "", __TIME__, __DATE__);
-#if 0 && defined(LCF_DSECS)
-	printf("MAX_IMAGES = %d\n", MAX_IMAGES);
-#endif
+        if (verbose>0) printf("Released %s%s and compiled at %s on %s%s\n",
+	        VERSION_DATE, comma ? "," : "", __TIME__, __DATE__, semi);
         printf("\n");
         if (version) return 0;
     }
@@ -670,7 +694,12 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 	die ("Only one of '-g', '-l', or '-L' may be specified");
 
     if (verbose > 0) errstd = stdout;
+    if (tell_param) tell_early = strcasecmp(tell_param, "chrul")
+    				&& strcasecmp(tell_param, "ebda");
     if (eflag) do_bitmap_edit(bitmap_file);
+    if (tell_param && tell_early) probe_tell(tell_param);
+    if (instmbr) do_install_mbr(act1, act2);
+    
 
     fd = cfg_open(config_file);
     more = cfg_parse(cf_options);
@@ -680,9 +709,9 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
     else nowarn = temp;
 
     if (verbose>=6) printf("main: cfg_parse returns %d\n", more);
-    if (tell_param) probe_tell(tell_param);
-    if (instmbr) do_install_mbr(act1, act2);
-    
+
+    if (tell_param && !tell_early) probe_tell(tell_param);
+
     if (!nowarn) {
 	if (fstat(fd,&st) < 0) {
 	    fprintf(errstd,"fstat %s: %s\n",config_file,strerror(errno));
@@ -695,19 +724,7 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 	    else if (st.st_mode & (S_IWGRP | S_IWOTH))
 		    fprintf(errstd,"Warning: %s should be writable only for "
 		      "root\n",config_file);
-#if 0
-		else {
-		    char *p = cfg_get_strg(cf_all,"password");
-		    char *pp = cfg_get_strg(cf_options,"password");
-		    
-		    if ( ((p && *p) || (pp && *pp))  &&
-			     (st.st_mode & (S_IRGRP | S_IROTH)) )
-			fprintf(errstd,"Warning: %s should be readable only "
-			  "for root if using PASSWORD\n",config_file);
-		}
-#else
 	    config_read = !!(st.st_mode & (S_IRGRP | S_IROTH));
-#endif
 	}
     }
     preload_dev_cache();
@@ -733,12 +750,10 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 	else if (geometric==AD_LINEAR) linear = 1;
 	else if (geometric==AD_GEOMETRIC) geom = 1;
     }    
-#ifdef LCF_LBA32
     if (geom+linear+lba32 == 0) {
 	if (!nowarn && verbose) fprintf(errstd,"Warning: LBA32 addressing assumed\n");
 	lba32 = 1;
     }
-#endif
     if (linear && !nowarn) fprintf(errstd,
     	"\nWarning:  LINEAR is deprecated in favor of LBA32:  LINEAR specifies 24-bit\n"
 	"  disk addresses below the 1024 cylinder limit; LBA32 specifies 32-bit disk\n"
@@ -746,32 +761,44 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 	"  use LINEAR only if you are aware of its limitations.\n\n");
     
     if (identify) identify_image(identify,ident_opt);
+    
+    if ( (param = cfg_get_strg(cf_options,"bios-passes-dl")) ) {
+	if (strchr("YyTt1", *param)) bios_passes_dl = DL_GOOD;
+	if (strchr("NnFf0", *param)) bios_passes_dl = DL_BAD;
+    }
+    if (bios_passes_dl == DL_NOT_SET) 	check_bios();	/* in probe.c */
+
+    
+
+    if (uninstall)
+	bsect_uninstall(uninst_dev ? uninst_dev : cfg_get_strg(cf_options,
+	      "boot"),cfg_get_strg(cf_options,"backup"),validate);
+    if (!nowarn && compact && (linear || lba32))
+	fprintf(errstd,"Warning: COMPACT may conflict with %s on some "
+		"systems\n", lba32 ? "LBA32" : "LINEAR");
+    if (reboot_arg) {
+	map_patch_first(cfg_get_strg(cf_options,"map") ? cfg_get_strg(
+	      cf_options,"map") : MAP_FILE, reboot_arg);
+/*	sync();	*/
+	exit(0);
+    }
+    if (argc) usage(name);
+    geo_init(cfg_get_strg(cf_options,"disktab"));
+    if (query)
+	show_images(!cfg_get_strg(cf_options,"map") ? MAP_FILE :
+	      cfg_get_strg(cf_options,"map"));
+
+
+/*************************************************/
+/*  Doing a real install (may be test mode)      */
+/*************************************************/
 
 /* test for a RAID installation */
-    raid_offset = raid_setup();
-    if (verbose >= 2) {
-	printf("raid_setup returns offset = %08lX  ndisk = %d\n", raid_offset, ndisk);
-	dump_serial_nos();    
-    }
-
-
-	if (uninstall)
-	    bsect_uninstall(uninst_dev ? uninst_dev : cfg_get_strg(cf_options,
-	      "boot"),cfg_get_strg(cf_options,"backup"),validate);
-	if (!nowarn && compact && (linear || lba32))
-	    fprintf(errstd,"Warning: COMPACT may conflict with %s on some "
-		"systems\n", lba32 ? "LBA32" : "LINEAR");
-	if (reboot_arg) {
-	    map_patch_first(cfg_get_strg(cf_options,"map") ? cfg_get_strg(
-	      cf_options,"map") : MAP_FILE, reboot_arg);
-	    sync();
-	    exit(0);
+	raid_offset = raid_setup();
+	if (verbose >= 2) {
+	    printf("raid_setup returns offset = %08lX  ndisk = %d\n", raid_offset, ndisk);
+	    dump_serial_nos();    
 	}
-	if (argc) usage(name);
-	geo_init(cfg_get_strg(cf_options,"disktab"));
-	if (query)
-	    show_images(!cfg_get_strg(cf_options,"map") ? MAP_FILE :
-	      cfg_get_strg(cf_options,"map"));
 
 	if (verbose >=2 && do_md_install)
 	    printf("raid flags: at bsect_open  0x%02X\n", raid_flags);
@@ -790,8 +817,11 @@ fprintf(errstd,"REBOOT=\"%s\"\n", reboot_arg);
 	    cfg_init(cf_top);
 	    if (cfg_parse(cf_top)) cfg_error("Syntax error");
 	}
-	if (!bsect_number())
-	    die("No images have been defined or default image doesn't exist.");
+	
+	temp = bsect_number();
+	if (temp==0) die("No images have been defined.");
+	else if (temp<0) die("Default image doesn't exist.");
+	
 	check_fallback();
 	
 	if (verbose>=2) dump_serial_nos();
