@@ -60,6 +60,7 @@ source directory.
 
 #define MAJMIN_RAM	0x101 /* RAM disk */
 #define MAJOR_FD	2 /* floppy disks */
+#define MAJOR_HPT370_SORT  MAJOR_FD	/* 114->2 for sorting HPT370 controller */
 #define MAJOR_HD	3 /* IDE-type hard disks */
 #define MAJOR_SD	8 /* SCSI disks 0-15 */
 #define MAJOR_MD	9 /* multi-disk RAID sets */
@@ -77,8 +78,9 @@ source directory.
 #define MAJOR_IDE6	57 /* IDE on sixth interface */
 #define MAJOR_LVM	58 /* Logical Volume Manager block device */
 #define MAJOR_EXPR	60 /* Experimental devices 60..63 */
-#define MAJOR_FL	62 /* M-Systems Disk-On-Chip 2000 ***experimental*** */
+/*#define MAJOR_FL	62 / M-Systems Disk-On-Chip 2000 ***experimental*** */
 
+#define MAJOR_SD_SORT	64 /*** MAJOR_SD converted to this for sorting ***/
 #define MAJOR_SD2	65 /* SCSI disks 16-31 */
 #define MAJOR_SD3	66 /* SCSI disks 32-47 */
 #define MAJOR_SD4	67 /* SCSI disks 48-63 */
@@ -105,6 +107,7 @@ source directory.
 #define MAX_DESCR_SECTORS  12  /* upper limit on MAX_DESCR_SECS */
 
 #ifdef LCF_PASS160
+#undef SHS_PASSWORDS
 #define SHS_PASSWORDS		/* use this one if SHS passwords are in use */
 #define MAX_PW_CRC	5  	/* max # of longwords in password digest */
 #define PW_FILE_SUFFIX ".shs"  	/* suffix for the file that saves password digest */
@@ -114,12 +117,6 @@ source directory.
 #define PW_FILE_SUFFIX ".crc"  	/* suffix for the file that saves password CRC's */
 #endif
 
-#ifndef MAX_SECONDARY_FILESIZE
-#define MAX_SECONDARY_FILESIZE 10240
-#endif
-
-#define MAX_SECONDARY ((MAX_SECONDARY_FILESIZE+SECTOR_SIZE-1)/SECTOR_SIZE)
-
 #ifdef LCF_DSECS
 #define MAX_DESCR_SECS LCF_DSECS
 #else
@@ -136,10 +133,6 @@ source directory.
 #define BLOCK_SIZE	1024 /* disk block size */
 #endif
 
-#if (MAX_SECONDARY_FILESIZE > MAX_SECONDARY*SECTOR_SIZE)
-#error "Largest secondary loader exceeds sector addresses reserved."
-#endif
-
 #define PARTITION_ENTRY	   16	/* size of one partition entry */
 #define PARTITION_ENTRIES  4    /* number of partition entries */
 #define PART_TABLE_SIZE    (PARTITION_ENTRY*PARTITION_ENTRIES)
@@ -149,15 +142,18 @@ source directory.
 
 #define P_MASK(x) ((~has_partitions(x))&0xFFFF)
 #define D_MASK(x) (has_partitions(x))
-#define PART_MAX    4 /* biggest primary partition number */
+#define PART_MAX PARTITION_ENTRIES /* biggest primary partition number */
+#define PART_MAX_MAX	63	/* max. partition number (on IDE disks) */
 
 #define MAX_BOOT_SIZE	0x1b6	/* (leave some space for NT's and DR DOS' dirty
 				   hacks) scream if the boot sector gets any
-				   bigger */
+				   bigger -- (22.5 - we now use those hacks) */
 
 #define BOOT_SIGNATURE	0xaa55	/* boot signature */
 #define BOOT_SIG_OFFSET	510	/* boot signature offset */
 #define MAGIC_SERIAL	0xC9CF	/* LILO installed serial number */
+#define PRIME		271	/* prime number used for serial no generation */
+#define SMALL_PRIME     17	/* another prime, but a small one */
 #define MAX_BIOS_DEVICES  16	/* max hard disk devices */
 #define MAX_RAID	30	/* max number of RAID disks in a set */
 #if 0
@@ -193,6 +189,8 @@ source directory.
 #define STAGE_SECOND	2	/* second stage loader code */
 #define	STAGE_CHAIN	0x10	/* chain loader code */
 #define STAGE_DRIVE_MAP	0x11	/* chain loader drive mapper */
+#define STAGE_MBR	0x12	/* mbr loader */
+#define STAGE_MBR2	0x13	/* mbr2 loader (extended) */
 #define STAGE_FLAG_SERIAL  0x0100	/* boot loader supports serial i/o */
 #define STAGE_FLAG_MENU	0x0200	/* boot loader supports menu interface */
 #define STAGE_FLAG_BMP4	0x0400	/* boot loader supports 640x480x4 bitmaps */
@@ -200,11 +198,14 @@ source directory.
 #define SETUPSECS	4	/* nr of setup sectors */
 #define VSS_NUM		497	/* address where variable setup size is
 				   stored */
+#define VGA_SET		506	/* address of initial kernel VGA mode */
 #define MAX_SETUPSECS	31	/* maximum number of sectors in kernel
 				   setup code (+ bootsect) */
 #define MAX_KERNEL_SECS	1024	/* absolute maximum kernel size */
 #define SPECIAL_SECTORS	2	/* special sectors (don't compact) at beginning
 				   of map sections */
+#define SPECIAL_BOOTSECT 1	/* INITSEG bootsect.S bootloader at beginning
+				   of all kernels, ahead of SETUPSECS */
 
 #define LINEAR_FLAG	0x40	/* mark linear address */
 #define LBA32_FLAG      0x20    /* mark lba 32-bit address */
@@ -238,11 +239,21 @@ source directory.
 #define FLAG_PROMPT	1	/* always issue boot: prompt */
 #define FLAG_RAID	2	/* one boot record of many */
 #define FLAG_RAID_DEFEAT  4	/* defeat finding this RAID boot record */
+#if 0
 #define FLAG_RAID_NOWRITE 8	/* defeat RAID writeback of command line */
+#endif
 #define FLAG_NOBD	16	/* defeat BIOS data collection at boot time */
 #define FLAG_LARGEMEM	32	/* BIOS has MoveExtMemBlk support for 386 */
 #define FLAG_MAP_ON_BOOT 64	/* map file is on the boot device */
+#define FLAG_BD_OKAY	128	/* BIOS data collection known to work */
+#ifdef FLAG_RAID_NOWRITE
 #define FLAG_SAVE (~(FLAG_RAID|FLAG_RAID_DEFEAT|FLAG_RAID_NOWRITE))  /* All but raid flags */
+#else
+#define FLAG_SAVE (~(FLAG_RAID|FLAG_RAID_DEFEAT))  /* All but raid flags */
+#endif
+
+/* these are the second-stage specific flags */
+#define FLAG2_EL_TORITO	2	/* El Torito format bootable CD */
 
 /* these are the descriptor flags */
 #define FLAG_VGA	1	/* override VGA mode */
@@ -250,7 +261,7 @@ source directory.
 #define FLAG_LOCK	4	/* lock on target */
 #define FLAG_MODKRN	8	/* modern kernel with high memory support */
 #define FLAG_KERNEL	16	/* image is a kernel */
-#define FLAG_SINGLE	32	/* single key activation */
+#define FLAG_TOOBIG	32	/* initrd so big, kernel may overwrite */
 #define FLAG_FALLBACK	64	/* fallback command line exists */
 #define FLAG_PASSWORD	128	/* this image requires a password */
 #define FLAG_LOADHI	256	/* this kernel loads high (>=1Mb) */
@@ -259,6 +270,8 @@ source directory.
 #define FLAG_VMWARN	1024	/* warn on virtual boot */
 #define FLAG_VMDEFAULT	2048	/* this is the default vitual load */
 #endif
+#define FLAG_SINGLE	4096	/* single key activation */
+
 
 #define VGA_NOCOVR	0x8000	/* VGA setting not overridden on command line */
 
@@ -306,8 +319,6 @@ INITSEG   = DEF_INITSEG			! we move boot here - out of the way
 SETUPSEG  = DEF_SETUPSEG		! setup starts here
 SYSSEG    = DEF_SYSSEG			! system loaded at 0x10000 (65536).
 
-MAX_SECONDARY_asm1 = MAX_SECONDARY_FILESIZE + SECTOR_SIZE - 1
-MAX_SECONDARY_asm = MAX_SECONDARY_asm1 / SECTOR_SIZE	! define for the assembler
 MAX_DESCR_SECS_asm = MAX_DESCR_SECS	! **
 MAX_DESCR_SECTORS_asm = MAX_DESCR_SECTORS ! **
 MAX_IMAGE_NAME_asm = MAX_IMAGE_NAME	! **
@@ -318,13 +329,6 @@ MAX_BIOS_DEVICES_asm = MAX_BIOS_DEVICES	! **
 MAX_RAID_DEVICES_asm = MAX_RAID_DEVICES	! **
 
 
-MAPcode   = MAX_SECONDARY_asm*0x200	! size of secondary code
-MAP	  = MAPcode+0x200		! plus space for cmdline input
-DFLCMD	  = MAP+0x200			! default command line
-DESCR	  = DFLCMD+0x200			! descriptor load area offset
-KEYTABLE  = DESCR+0x200*MAX_DESCR_SECS_asm	     ! 0x2600	! default command line
-DATAEND	  = KEYTABLE+0x200   ! 0x2A00	! PARAMLINE or Load-Low area
-
 STACKSEG  = 0x9000	     ! MUST == INITSEG for kernel 2.0.36 (and others?)
 SETUP_STACKSIZE = 2048		! stacksize for kernel setup.S
 
@@ -337,12 +341,8 @@ SETUP_STACKSIZE = 2048		! stacksize for kernel setup.S
 #ifdef LILO_ASM
 
 STACK	  = 2048		! amount of stack space to reserve
-PARMLINE  = DATAEND		! 0x2A00	! parameter line construction area off.
-
 SSDIFF	= 0
 
-!SLA_SIZE  = 0x9E00		! setup load area size
-				! as the max. heap pointer in newer kernels
 BOOTSECT  = 0x200		! kernel bootsect.S
 
 #define SETUP_STACK_DYN  PARMLINE
@@ -405,19 +405,20 @@ CL_RAMDISK_MAX	= CL_POINTER+4		! ramdisk_max; header version 0x0203
    don't compile without them ...
  */
 
-#ifndef LCF_M386
-# define LCF_M386
-#endif
-#ifndef LCF_VARSETUP
-# define LCF_VARSETUP
-#endif
 #ifndef LCF_UNIFY
 # define LCF_UNIFY
 #endif
 #ifndef LCF_BUILTIN
 # define LCF_BUILTIN
 #endif
-
+#ifndef LCF_FIRST6
+# define LCF_FIRST6
+#endif
+#if VERSION_MINOR >= 50
+# define BETA_TEST 1
+#else
+# define BETA_TEST 0
+#endif
 
 #endif
 
