@@ -93,14 +93,7 @@ void map_create(char *name)
 	    printf("map_create:  boot=%04X  map=%04X\n", 
 					boot_dev_nr, (int)st.st_dev);
 	if (boot_dev_nr != st.st_dev  &&  extra != X_MBR_ONLY) {
-#if 1
 	    die("map file must be on the boot RAID partition");
-#else
-	    raid_flags |= FLAG_RAID_NOWRITE;
-	    fprintf(errstd,
-	    	"Error: map file must be on the boot RAID partition\n"
-	    	"    This installation provides no redundancy\n"	);
-#endif
 	}
     }
 
@@ -230,7 +223,7 @@ static void map_compact(int dont_compact)
 	maxcount = lba32 ? 3 : maxcount;
 #endif
     walk = map;
-    for (count = 0; walk && count <= dont_compact; count++) walk = walk->next;
+    for (count = 0; walk && count < dont_compact; count++) walk = walk->next;
     offset = 0;
     noffset = 0;
     while (walk && walk->next) {
@@ -286,7 +279,7 @@ static void map_alloc_page(int offset,SECTOR_ADDR *addr)
 
     if ((here = lseek(map_file,offset,SEEK_CUR)) < 0) pdie("lseek map file");
     if (write(map_file,"",1) != 1) pdie("write map file");
-    if (fsync(map_file)) pdie("fsync map file");
+    if (fdatasync(map_file)) pdie("fdatasync map file");
     if (!geo_comp_addr(&map_geo,here,addr))
 	die("Hole found in map file (alloc_page)");
     if (lseek(map_file,-offset-1,SEEK_CUR) < 0) pdie("lseek map file");
@@ -342,21 +335,83 @@ int map_end_section(SECTOR_ADDR *addr,int dont_compact)
     return sectors;
 }
 
+#ifdef LCF_FIRST6
+static long sa6_from_sa(SECTOR_ADDR6 *sa6, SECTOR_ADDR *sa)
+{
+    static unsigned char hinib = 0;
+    int count;
+    int flags;
+    long sector;
+    
+    sa6->device = sa->device & DEV_MASK;
+    flags = sa6->flags = sa->device & ~DEV_MASK;
+    count = sa->num_sect;
+    
+    if ((flags & (LBA32_FLAG|LINEAR_FLAG)) == 0) {
+/* pure geometric addressing */
+	sector = *(unsigned long*)&(sa->sector);
+    }
+    else {
+	if (flags & LBA32_FLAG) {
+/* pure lba32 addressing */
+	    if (flags & LBA32_NOCOUNT) {
+		hinib = count;
+		count = 1;
+	    }
+	sector = hinib;
+	} /* linear addressing */
+	else  sector = 0;
 
+	sector = ((sector<<8 | sa->head)<<8 | sa->track)<<8 | sa->sector;
+    }
+    sa6->sector = sector;
+    
+    return sector;
+}
+
+#endif
+
+
+#ifdef LCF_FIRST6
+int map_write(SECTOR_ADDR *list,int max_len,int terminate,int sa6)
+#else
 int map_write(SECTOR_ADDR *list,int max_len,int terminate)
+#endif
 {
     MAP_ENTRY *walk,*next;
     int sectors;
+#ifdef LCF_FIRST6
+    SECTOR_ADDR6 sa6tem, *list6 = (void*)list;
+    unsigned long *list4 = (void*)list;
+#endif
 
     sectors = 0;
     for (walk = map; walk; walk = next) {
 	next = walk->next;
 	if (--max_len < (terminate ? 1 : 0)) die("Map segment is too big.");
+#ifdef LCF_FIRST6
+	if (sa6) {
+	    (void)sa6_from_sa(&sa6tem, &(walk->addr));
+	    if (sa6==2) *list4++ = sa6tem.sector;
+	    else *list6++ = sa6tem;
+	}
+	else
+#endif
 	*list++ = walk->addr;
+
 	free(walk);
 	sectors++;
     }
-    if (terminate) memset(list,0,sizeof(SECTOR_ADDR));
+    
+    if (terminate) {
+#ifdef LCF_FIRST6
+	if (sa6==2) *list4 = 0;
+	else if (sa6) memset(list6, 0, sizeof(SECTOR_ADDR6));
+	else
+#endif
+	memset(list,0,sizeof(SECTOR_ADDR));
+    }	
+
     return sectors;
 }
 
@@ -415,3 +470,5 @@ off_t map_insert_data(unsigned char *data, int size)
     
     return here;
 }
+
+
