@@ -36,20 +36,34 @@ source directory.
 
 
        
-       _syscall5(int,  _llseek,  uint,  fd, ulong, hi, ulong, lo,
-       loff_t *, res, uint, wh);
-
+       _syscall5(int,  _llseek,  unsigned int,  fd, unsigned long, hi,
+           unsigned long, lo, loff_t *, res, unsigned int, wh);
        int _llseek(unsigned int fd,  unsigned  long  offset_high,
-       unsigned  long  offset_low,  loff_t * result, unsigned int
-       whence);
-       
+           unsigned  long  offset_low,  loff_t * result, unsigned int whence);
+
        loff_t llseek(unsigned int fd, loff_t offs, unsigned int whence)
        { loff_t res;
-       	return _llseek(fd, offs>>32, offs, &res, whence) < 0  ?
+       	   return _llseek(fd, offs>>32, offs, &res, whence) < 0  ?
        			 (loff_t)(-1) : res;
        }
 
-
+static
+int anywhere(void *buf, unsigned char *str)
+{
+    int k, n;
+    void *s;
+    
+    k = strlen(str);
+    n = SECTOR_SIZE-k;
+    s = memchr(buf, *str, n);
+    while(s) {
+	if (!strncmp(s, str, k)) return 1;
+	s++;
+	n = SECTOR_SIZE - k - (int)(s-buf);
+	s = memchr(s, *str, n);
+    }
+    return 0;
+}
 
 
 
@@ -67,21 +81,21 @@ int part_nowrite(char* device)
     int fd;
     BOOT_SECTOR bs;
     
-    int ret=0;	/* say ok, unless we recognize a problem partition */
-
+    int ret=PTW_OKAY;	/* say ok, unless we recognize a problem partition */
+if ( !(do_md_install && extra==X_MBR_ONLY) ) {
     if ((fd = open(device, O_RDONLY)) < 0) pdie("part_nowrite check:");
     if (read(fd, bs.sector, sizeof(bs)) != SECTOR_SIZE) pdie("part_nowrite: read:");
     
 /* check for XFS */
-    if (!strncmp("XFSB", bs.sector, 4)) ret=3;
+    if (!strncmp("XFSB", bs.sector, 4)) ret=PTW_XFS;
     
 /* check for NTFS */
     else if (	!strncmp("NTFS", bs.par_d.system, 4)
-		|| strstr(bs.sector,"NTLDR")  ) ret=2;
+		|| anywhere(bs.sector,"NTLDR")  ) ret=PTW_NTFS;
 
 /* check for HPFS */
     else if (	!strncmp("OS2", bs.par_d.system, 3)
-		|| strstr(bs.sector,"OS2LDR")  ) ret=2;
+		|| anywhere(bs.sector,"OS2LDR")  ) ret=PTW_OS2;
 
 /* check for DOS FAT */
     else if (
@@ -91,7 +105,13 @@ int part_nowrite(char* device)
     /* sectors_per_cluster is a power of 2, meaning only 1 bit is on */
 	&& bs.par_d.bpb.sectors_per_cluster
 	&& (bs.par_d.bpb.sectors_per_cluster & (bs.par_d.bpb.sectors_per_cluster-1))==0
-				) ret=1;
+				) {
+		ret=PTW_DOS;
+#if 0
+/* this, it turns out is from Windows 98, so no caution here on NT */
+    		if (anywhere(bs.sector,"WINBOOT SYS")) ret+=PTW_NTFS;
+#endif
+    }
     
 /* check for SWAP -- last check, as 'bs' is overwritten */
     else if (*(long*)bs.sector == 0xFFFFFFFEUL) {
@@ -99,19 +119,16 @@ int part_nowrite(char* device)
 	    pdie("part_nowrite lseek:");
 	if (SECTOR_SIZE != read(fd, bs.sector, sizeof(bs)) ) pdie("part_nowrite swap check:");
 	if (!strncmp(bs.sector+SECTOR_SIZE-10,"SWAPSPACE2",10)
-	    || !strncmp(bs.sector+SECTOR_SIZE-10,"SWAP-SPACE",10) ) ret=3;
+	    || !strncmp(bs.sector+SECTOR_SIZE-10,"SWAP-SPACE",10) ) ret=PTW_SWAP;
     }
 
 /* didn't recognize the superblock type, so assume it is okay */    
-    else ret=0;
+    else ret=PTW_OKAY;
     
     close(fd);
 
-/* check for override (-F) command line flag */
-    if (ret>1 && force_fs) {
-	fprintf(errstd, "WARNING: '-F' override used. Filesystem on  %s  may be destroyed.\n", device);
-	ret=0;
-    }
+} /* raid install with X_MBR_ONLY in use */
+    if (verbose>=6) printf("part_nowrite: %d\n", ret);    
     
     return ret;
 }
@@ -125,8 +142,9 @@ void part_verify(int dev_nr,int type)
     int fd, bck_file, part, size, lin_3d, cyl;
     unsigned long second, base;
     struct partition part_table[PART_MAX];
-    int mask, i, pe, Linux, dos;
+    int mask, i, pe, Linux, dos, mbr;
     unsigned short boot_sig;
+    BOOT_PARAMS_1 bs;
     
     if (!has_partitions(dev_nr) || !(mask = P_MASK(dev_nr)) || !(dev_nr & mask)
 #if 0
@@ -139,6 +157,14 @@ void part_verify(int dev_nr,int type)
     fd = dev_open(&dev,dev_nr & ~mask,cfg_get_flag(cf_options,"fix-table")
       && !test ? O_RDWR : O_RDONLY);
     part = (pe = dev_nr & mask)-1;
+#if 1
+    if (type) {
+	if (lseek(fd, 0L, SEEK_SET) != 0 ||
+	    read(fd, &bs, sizeof(bs)) != sizeof(bs) ) pdie("bs read");
+	if (*(long*)bs.signature==EX_MAG_HL) mbr = bs.stage;
+	else mbr = STAGE_MBR;
+    } else mbr = STAGE_MBR;
+#endif
     if (lseek(fd, PART_TABLE_OFFSET, SEEK_SET) < 0) pdie("lseek partition table");
     if (!(size = read(fd,(char *) part_table, sizeof(struct partition)*
       PART_MAX))) die("Short read on partition table");
@@ -167,7 +193,12 @@ void part_verify(int dev_nr,int type)
         i++;
         part=0;
     }
-
+#if 1
+    if (type && pe>0 && pe<=(mbr==STAGE_MBR2?63:PART_MAX)
+    	     && !(part_table[part].boot_ind&0x80) && !nowarn)
+	fprintf(errstd,"Warning: Partition %d on %s is not marked Active.\n",
+		pe, dev.name);
+#endif
     i = part_table[part].sys_ind;
 
     Linux =   i == PART_LINUX_MINIX ||
@@ -184,21 +215,18 @@ void part_verify(int dev_nr,int type)
 
     if (type && !Linux) {
 	fflush(stdout);
-#if 0
-	if (!dos || !nowarn)
-	    fprintf(errstd,"Device 0x%04X: Partition type 0x%02X does not seem "
-	      "suitable for a LILO boot sector\n",dev_nr,part_table[part].sys_ind);
-	if (dos && !nowarn)
-	    fprintf(errstd,"but DOS/Windows/OS2 FAT partitions are now usable.\n");
-#else
-	if (!nowarn) fprintf(errstd,"Warning: partition type 0x%02X"
-		" on device 0x%04X is an odd place for a boot\n  sector.",
+	if (!nowarn) fprintf(errstd,"Warning: partition type 0x%02X"" on device 0x%04X is a dangerous place for\n"
+							"    a boot sector.",
 			part_table[part].sys_ind, dev_nr);
 	if (dos && !nowarn) fprintf(errstd,"  A DOS/Windows/OS2 system may be rendered unbootable."
 		"\n  The backup copy of this boot sector should be retained.");
 	if (!nowarn) fprintf(errstd,"\n");
+#if 0
+	if (!dos && !cfg_get_flag(cf_options,"ignore-table"))
+	    die("You may proceed by using either '-P ignore' or 'ignore-table'");
+#else
+	if (!yesno("\nProceed? ", 0)) exit(0);
 #endif
-	if (!dos && !cfg_get_flag(cf_options,"ignore-table")) exit(1);
     }
     cyl = part_table[part].cyl+((part_table[part].sector >> 6) << 8);
     lin_3d = (part_table[part].sector & 63)-1+(part_table[part].head+
@@ -245,7 +273,7 @@ void part_verify(int dev_nr,int type)
 		printf("Backup copy of partition table in %s\n",backup_file);
 	    printf("Writing modified partition table to device 0x%04X\n",
 	      dev_nr & ~mask);
-	    if (lseek(fd,PART_TABLE_OFFSET,0) < 0)
+	    if (lseek(fd,PART_TABLE_OFFSET,SEEK_SET) < 0)
 		pdie("lseek partition table");
 	    if (!(size = write(fd,(char *) part_table,sizeof(struct partition)*
 	      PART_MAX))) die("Short write on partition table");
@@ -387,7 +415,8 @@ void do_cr_auto(void)
     struct partition part_table[PART_MAX];
 
     if (autoauto) has_partition = 0;
-    other = cfg_get_strg(cf_top, "other");
+    other = identify ? cfg_get_strg(cf_identify, "other")
+		     : cfg_get_strg(cf_top, "other");
     if (verbose > 4) printf("do_cr_auto: other=%s has_partition=%d\n",
         other, has_partition);
 #if 0
@@ -544,104 +573,246 @@ void preload_types(void)
 #define PART_INACTIVE	0
 
 
-void do_activate(char *where, char *which)
+void do_activate(char *part, char *which)
 {
+#if 1
+    int part_max, count, number, fd;
+    struct partition pt [PART_MAX_MAX+1];
+    long long daddr [PART_MAX_MAX+1];
+    int modify=0;
+    
+    part_max = read_partitions(part, extended_pt ? PART_MAX_MAX : 0,
+    					NULL, pt, daddr);
+/*    printf("part_max=%d\n", part_max); */
+    if (!which) {	/* one argument: display active partition */
+	for (count=0; count < part_max; count++) {
+	    if (pt[count].boot_ind) {
+		printf("%s%d\n",part,count+1);
+		exit(0);
+	    }
+	}
+	printf("No active partition found on %s\n",part);
+	exit(0);
+    }
+    number = to_number(which);
+    if (number < 0 || number > part_max)
+	die("%s: not a valid partition number (1-%d)",which,part_max);
+
+    if (number && !pt[number-1].sys_ind) die("Cannot activate an empty partition");
+    number--;	/* we are zero-based from here on */
+
+    if ((fd = open(part, O_RDWR)) < 0)
+	die("open %s: %s",part,strerror(errno));
+    for (count=0; count<part_max; count++) {
+    	unsigned char flag = count==number ? PART_ACTIVE : PART_INACTIVE;
+    	if (pt[count].sys_ind && pt[count].boot_ind != flag) {
+    	    pt[count].boot_ind = flag;
+	    printf("pt[%d] -> %2x\n", count+1, (int)flag);
+	    if (llseek(fd, daddr[count], SEEK_SET) < 0) die("PT llseek failed");
+	    if (!test)
+	    if (write(fd, &pt[count], sizeof(pt[0])) != sizeof(pt[0]) )
+		die("PT write failure");
+	    modify++;
+    	}
+    }
+    close(fd);
+    if (modify)
+	printf("The partition table has%s been updated.\n", test ? " *NOT*" : "");
+    else
+	printf("No partition table modifications are needed.\n");
+#else
     struct stat st;
     int fd,number,count;
     unsigned char flag, ptype;
 
-    if ((fd = open(where, !which ? O_RDONLY : O_RDWR)) < 0)
-	die("open %s: %s",where,strerror(errno));
-    if (fstat(fd,&st) < 0) die("stat %s: %s",where,strerror(errno));
-    if (!S_ISBLK(st.st_mode)) die("%s: not a block device",where);
+    if ((fd = open(part, !which ? O_RDONLY : O_RDWR)) < 0)
+	die("open %s: %s",part,strerror(errno));
+    if (fstat(fd,&st) < 0) die("stat %s: %s",part,strerror(errno));
+    if (!S_ISBLK(st.st_mode)) die("%s: not a block device",part);
     if (verbose >= 1) {
        printf("st.st_dev = %04X, st.st_rdev = %04X\n",
        				(int)st.st_dev, (int)st.st_rdev);
     }
     if ((st.st_rdev & has_partitions(st.st_rdev)) != st.st_rdev)
-        die("%s is not a master device with a primary partition table", where);
+        die("%s is not a master device with a primary partition table", part);
     if (!which) {	/* one argument: display active partition */
 	for (count = 1; count <= PART_NUM; count++) {
-	    if (lseek(fd,PART_BEGIN+(count-1)*PART_SIZE,0) < 0)
+	    if (lseek(fd,PART_BEGIN+(count-1)*PART_SIZE,SEEK_SET) < 0)
 		die("lseek: %s",strerror(errno));
 	    if (read(fd,&flag,1) != 1) die("read: %s",strerror(errno));
 	    if (flag == PART_ACTIVE) {
-		printf("%s%d\n",where,count);
+		printf("%s%d\n",part,count);
 		exit(0);
 	    }
 	}
-	die("No active partition found on %s",where);
+	die("No active partition found on %s",part);
     }
     number = to_number(which);
     if (number < 0 || number > 4)
 	die("%s: not a valid partition number (1-4)",which);
     for (count = 1; count <= PART_NUM; count++) {
-	if (lseek(fd,PART_BEGIN+(count-1)*PART_SIZE+4,0) < 0)
+	if (lseek(fd,PART_BEGIN+(count-1)*PART_SIZE+4,SEEK_SET) < 0)
 	    die("lseek: %s",strerror(errno));
 	if (read(fd,&ptype,1) != 1) die("read: %s",strerror(errno));
 	if (count == number && ptype==0) die("Cannot activate an empty partition");
     }
     if (test) {
-        printf("The partition table of  %s  has *NOT* been updated\n",where);
+        printf("The partition table of  %s  has *NOT* been updated\n",part);
     }
     else for (count = 1; count <= PART_NUM; count++) {
-	if (lseek(fd,PART_BEGIN+(count-1)*PART_SIZE,0) < 0)
+	if (lseek(fd,PART_BEGIN+(count-1)*PART_SIZE,SEEK_SET) < 0)
 	    die("lseek: %s",strerror(errno));
 	flag = count == number ? PART_ACTIVE : PART_INACTIVE;
 	if (write(fd,&flag,1) != 1) die("write: %s",strerror(errno));
     }
+#endif
     exit(0);
 }
 
 
-void do_install_mbr(char *where, char *what)
+void do_install_mbr(char *part, char *what)
 {
     int fd, i;
 #ifndef LCF_BUILTIN
     int nfd;
 #endif
     struct stat st;
-    char buf[SECTOR_SIZE];
+    BOOT_SECTOR buf;
+    char *cp;
     
     if (!what) what = DFL_MBR;
-    if ((fd=open(where,O_RDWR)) < 0) die("Cannot open %s: %s", where,strerror(errno));
-    if (fstat(fd,&st) < 0) die("stat: %s : %s", where,strerror(errno));
-    if (!S_ISBLK(st.st_mode)) die("%s not a block device",where);
+    extended_pt |= !!strchr(what,'x') || !!strchr(what,'X') || !!strchr(what,'2');
+    if ((fd=open(part,O_RDWR)) < 0) die("Cannot open %s: %s", part,strerror(errno));
+    if (fstat(fd,&st) < 0) die("stat: %s : %s", part,strerror(errno));
+    if (!S_ISBLK(st.st_mode)) die("%s not a block device",part);
     if (st.st_rdev != (st.st_rdev & has_partitions(st.st_rdev)))
-	die("%s is not a master device with a primary parition table",where);
-    if (read(fd,buf,SECTOR_SIZE) != SECTOR_SIZE) die("read %s: %s",where, strerror(errno));
+	die("%s is not a master device with a primary parition table",part);
+    if (read(fd,&buf,SECTOR_SIZE) != SECTOR_SIZE) die("read %s: %s",part, strerror(errno));
+
+    cp = cfg_get_strg(cf_options,"force-backup");
+    i = (cp!=NULL);
+    if (!cp) cp = cfg_get_strg(cf_options,"backup");
+    make_backup(cp, i, &buf, st.st_rdev, part);
+    
 #ifndef LCF_BUILTIN    
     if ((nfd=open(what,O_RDONLY)) < 0) die("Cannot open %s: %s",what,strerror(errno));
     if (read(nfd,buf,MAX_BOOT_SIZE) != MAX_BOOT_SIZE) die("read %s: %s",what,strerror(errno));
 #else
-    memcpy(buf, Mbr.data, MAX_BOOT_SIZE);
+    memcpy(&buf, extended_pt ? Mbr2.data : Mbr.data, MAX_BOOT_SIZE);
 #endif    
-    *(unsigned short*)&buf[BOOT_SIG_OFFSET] = BOOT_SIGNATURE;
+    buf.boot.boot_ind = BOOT_SIGNATURE;
     if (zflag) {
-        char *p = buf+MAX_BOOT_SIZE;
-        for (i=0; i<8; i++) *p++ = 0;
-    } else if (*(int*)&buf[PART_TABLE_OFFSET - 6] == 0) {
+        buf.boot.mbz =
+        buf.boot.marker =
+        buf.boot.volume_id = 0;
+#if BETA_TEST || 1
+        if ((cp=cfg_get_strg(cf_options,RAID_EXTRA_BOOT))) {
+	    buf.boot.volume_id = strtoul(cp, NULL, 16);
+        }
+#endif
+    } else if (buf.boot.volume_id == 0) {
+#if 0
     	i = st.st_rdev;
-    	i %= 271;		/* modulo a prime number; eg, 2551, 9091 */
-    	i += 31;
+    	i %= PRIME;		/* modulo a prime number; eg, 2551, 9091 */
+    	i += SMALL_PRIME;
         srand(time(NULL));	/* seed the random number generator */
         while (i--) rand();
         *(int*)&buf[PART_TABLE_OFFSET - 6] = rand();  /* insert serial number */
         if (*(short*)&buf[PART_TABLE_OFFSET - 2] == 0)
             *(short*)&buf[PART_TABLE_OFFSET - 2] = MAGIC_SERIAL;
+#else
+	buf.boot.volume_id = new_serial(st.st_rdev);
+	buf.boot.marker = MAGIC_SERIAL;
+#endif
     }
     
-    if (lseek(fd,0,SEEK_SET) != 0) die("seek %s; %s", where, strerror(errno));
+    if (lseek(fd,0,SEEK_SET) != 0) die("seek %s; %s", part, strerror(errno));
     if (!test) {
-	if (write(fd,buf,SECTOR_SIZE) != SECTOR_SIZE)
-		die("write %s: %s",where,strerror(errno));
+	if (write(fd,&buf,SECTOR_SIZE) != SECTOR_SIZE)
+		die("write %s: %s",part,strerror(errno));
     }
     close(fd);
 #ifndef LCF_BUILTIN
     close(nfd);
 #endif
-    printf("The Master Boot Record of  %s  has %sbeen updated.\n", where, test ? "*NOT* " : "");
-    sync();
+    printf("The Master Boot Record of  %s  has %sbeen updated.\n", part, test ? "*NOT* " : "");
     exit(0);
 }
 
+
+
+/* partition table read */
+int read_partitions(char *part, int max, long *volid,
+		struct partition *p, long long *where)
+{
+    int fd, i;
+    unsigned int second, base;
+    unsigned short boot_sig;
+    struct partition pt[PART_MAX];
+    BOOT_PARAMS_1 hdr;
+    struct stat st;
+    long long daddr;
+
+    if ((fd=open(part,O_RDONLY))<0) die("Cannot open '%s'", part);
+    if (fstat(fd,&st)<0) die("Cannot fstat '%s'", part);
+    if (!S_ISBLK(st.st_mode)) die("Not a block device '%s'", part);
+    i = st.st_rdev;
+    if (!has_partitions(i) || (P_MASK(i)&i) )
+	die("Not a device with partitions '%s'", part);
+
+    if (read(fd, &hdr, sizeof(hdr)) != sizeof(hdr)) die("read header");
+    if (!strncmp(hdr.signature, "LILO", 4) && hdr.stage == STAGE_MBR2 &&
+	max == 0) max = PART_MAX_MAX;
+    else if (max == 0) max = PART_MAX;
+    if (lseek(fd, PART_TABLE_OFFSET, SEEK_SET)<0) die("lseek failed");
+    if (read(fd, pt, sizeof(pt)) != sizeof(pt)) die("read pt failed");
+    if ( read(fd, &boot_sig, sizeof(boot_sig)) != sizeof(boot_sig)  ||
+	boot_sig != BOOT_SIGNATURE ) die("read boot signature failed");
+    if (volid) {
+	if (lseek(fd, MAX_BOOT_SIZE+2, SEEK_SET)<0) die("lseek vol-ID failed");
+	if (read(fd, volid, sizeof(*volid)) != sizeof(*volid))
+	    die("read vol-ID failed");
+/*	printf(" vol-ID: %08X\n", second);	*/
+    }
+/*    printf("%s\n", phead); */
+    second=base=0;
+    if (max>=4)
+    for (i=0; i<PART_MAX; i++) {
+/*	print_pt(i+1, pt[i]); */
+	if (is_extd_part(pt[i].sys_ind)) {
+	    if (!base) base = pt[i].start_sect;
+	    else die("invalid partition table: second extended partition found");
+	}
+	if (where) *where++ = PART_TABLE_OFFSET + i*sizeof(*p);
+	*p++ = pt[i];
+    }
+
+    max -= (i=4);
+    
+    if (max>0)
+    while (base) {
+    	daddr = LLSECTORSIZE*(base+second) + PART_TABLE_OFFSET;
+        if (llseek(fd, daddr, SEEK_SET) < 0)
+            die("secondary llseek failed");
+	if (read(fd, pt, sizeof(pt)) != sizeof(pt)) die("secondary read pt failed");
+	if ( read(fd, &boot_sig, sizeof(boot_sig)) != sizeof(boot_sig)  ||
+	    boot_sig != BOOT_SIGNATURE ) die("read second boot signature failed");
+/*	print_pt(i++, pt[0]); */
+        if (is_extd_part(pt[1].sys_ind)) second=pt[1].start_sect;
+        else base = 0;
+        if (max-- > 0) {
+	    *p++ = pt[0];
+	    if (where) *where++ = daddr;
+	    i++;
+        }
+    }
+    if (max > 0) {
+	p->sys_ind = 0;
+	if (where) *where = 0;
+    }
+        
+    close(fd);
+    
+    return i;
+}
+
