@@ -10,7 +10,7 @@ source directory.
 
 /*#define DEBUG*/
 #define BITMAP 0	/* change to 1 when do_bitmap is filled in */
-#define VIDEO 1
+#define VOLID  1	/* change to 1 when do_volid is filled in */
 
 #include <stdlib.h>
 #include <unistd.h>
@@ -46,7 +46,7 @@ source directory.
 
 
 static union Buf {
-   unsigned char b[4*SECTOR_SIZE];
+   unsigned char b[5*SECTOR_SIZE];
    struct {
       short checksum[2];	/* prevent alignment on *4 boundary */
       char signature[4];
@@ -61,9 +61,31 @@ static union Buf {
       floppy_t f[4];
       hard_t d;
 /*      edd_t edd; */
-   } s;
+   } s4;
+   struct {
+      short checksum[2];	/* prevent alignment on *4 boundary */
+      char signature[4];
+      short version;
+      short length;
+      unsigned char disk;	/* device code of last good disk */
+      unsigned char vid, mflp, mhrd;
+      short floppy;		/* byte offset to floppy data    */
+      short hard;		/* byte offset to hard disk data */
+      short partitions;		/* byte offset to partition info */
+/* version 5 additions	*/
+      short equipment;		/* byte offset to the equipment information */
+      short video;		/* byte offset to the video information */
+   } s5;
 } buf;
-static int buf_valid = 0;
+
+static equip_t *eq;
+static video_t1 *v1;
+static video_t2 *v2;
+static video_t25 *v25;	/* extension for PROBE_VERSION 5 */
+static video_t3 *v3;
+
+static int video_36_bug;
+static int buf_valid = -1;
 static hard_t *hdp[16+1] =		/* pointers to all the hard disks */
 	{	NULL, NULL, NULL, NULL, 
 		NULL, NULL, NULL, NULL, 
@@ -77,35 +99,19 @@ static void do_help(void);
 static void do_geom(char *bios);
 static void do_geom_all(void);
 static void do_table(char *part);
-#if VIDEO
 static void do_video(void);
-#endif
+static void do_bios(void);
 #if BITMAP
 static void do_bitmap(char *file);
 #endif
+#if VOLID
+static void do_volid(void);
+#endif
 static char dev[] = "<device>";
 
 extern CHANGE_RULE *change_rules;	/* defined in partition.c */
 
 
-#if 0   /* def LCF_ALL_PARTITIONS */
-
-#define LLSECTORSIZE ((long long)SECTOR_SIZE)
-
-       _syscall5(int,  _llseek,  uint,  fd, ulong, hi, ulong, lo,
-       loff_t *, res, uint, wh);
-
-       int _llseek(unsigned int fd,  unsigned  long  offset_high,
-       unsigned  long  offset_low,  loff_t * result, unsigned int
-       whence);
-       
-       loff_t llseek(unsigned int fd, loff_t offs, unsigned int whence)
-       { loff_t res;
-       	return _llseek(fd, offs>>32, offs, &res, whence) < 0  ?
-       			 (loff_t)(-1) : res;
-       }
-#endif
-
 
 static
 struct Probes {
@@ -116,6 +122,7 @@ struct Probes {
 	}
 	list[] = {
 { "help",  do_help,  NULL,  "Print list of -T(ell) options"	},
+{ "bios",  do_bios,  NULL,  "State of DL as passed to boot loader"  },
 #if BITMAP
 { "bitmap=",do_bitmap,"<file>", "Display .bmp file X,Y/color/timer information"},
 #endif
@@ -124,8 +131,9 @@ struct Probes {
 { "geom=", do_geom,  "<bios>", "Geometry CHS data for BIOS code 0x80, etc." },
 { "geom" , do_geom_all, NULL, "Geometry for all BIOS drives" },
 { "table=", do_table, dev, "Partition table information for /dev/hda, etc."},
-#if VIDEO
 { "video", do_video, NULL,  "Graphic mode information" },
+#if VOLID
+{ "vol-ID", do_volid, NULL, "Volume ID check for uniqueness"},
 #endif
 { NULL,    NULL,     NULL,   NULL}
 	};
@@ -156,48 +164,110 @@ static struct partitions {
     { NULL, 0, 0 }   };
 
 static char phead[] = "\t\t Type  Boot      Start           End      Sector    #sectors";
-
+static int dirty = -1;	/* buffer is unread */
 
 /* load the low memory bios data area */
 /*  0 = no error, !0 = error on get */
-static int fetch(void)
+int fetch(void)
 {
     int fd;
     int got, get;
     int at = 0, seek = PROBESEG*16;
     
-    if (buf_valid) return 0;
+    if (buf_valid>=0) return buf_valid;
     
-    if ((fd=open("/dev/mem", O_RDONLY)) < 0) return 1;
-#if 0
-    while (at<seek) {
-	get = seek - at > sizeof(buf.b) ? sizeof(buf.b) : seek - at;
-	got = read(fd, &buf.b, get);
-	if (got != get) return 1;
-	at += got;
-    }
-#else
+    if ((fd=open("/dev/mem", O_RDONLY)) < 0) return buf_valid=1;
     at = lseek(fd, seek, SEEK_SET);
-    if (at != seek) return 1;
-#endif
+    if (at != seek) return buf_valid=1;
     get = sizeof(buf.b);
-    if (read(fd, &buf.b, get) != get) return 1;
+    if (read(fd, &buf.b, get) != get) return buf_valid=1;
     close(fd);
+    dirty = 0;	/* buffer is unmodified */
     
-    if (strncmp(buf.s.signature, PROBE_SIGNATURE,4)) return 2;
-/*    got = buf.s.version; */	/* quiet GCC */
-    if (buf.s.version < 3 ||
-        buf.s.version > (short)(PROBE_VERSION)) return 3;
-    got = buf.s.length;
-    if (got > sizeof(buf.b) || got < sizeof(buf.s)) return 4;
-    if (*(long*)buf.s.checksum != crc32((char*)&buf.s + 4, got-4, CRC_POLY1))
-    	return 5;
-    buf_valid = 1;
-    
-    return 0;
+    if (strncmp(buf.s5.signature, PROBE_SIGNATURE,4)) return buf_valid=2;
+/*    got = buf.s5.version; */	/* quiet GCC */
+    if (buf.s5.version < 3 ||
+        buf.s5.version > (short)(PROBE_VERSION)) return buf_valid=3;
+    got = buf.s5.length;
+    if (got > sizeof(buf.b) || got < sizeof(buf.s4)) return buf_valid=4;
+    if (*(long*)buf.s5.checksum != crc32((char*)&buf.s5 + 4, got-4, CRC_POLY1))
+    	return buf_valid=5;
+
+    if (buf.s5.version == 4) {
+	eq = (void*)&buf.s4.v.equipment;
+	v1 = (void*)&buf.s4.v.vid0F;
+	if (buf.s4.vid > 1) v2 = (void*)&buf.s4.v.vid12;
+	if (buf.s4.vid > 2) v3 = (void*)&buf.s4.v.vid4F00;
+    }
+    if (buf.s5.version >= 5) {
+	eq = (void*)&buf.b[buf.s5.equipment];
+	v1 = (void*)&buf.b[buf.s5.video];
+	if (buf.s5.vid > 1) {
+	    v2 = (void*)v1 + sizeof(*v1);
+	    v25 = (void*)v2 + sizeof(*v2);
+	}
+	if (buf.s5.vid > 2) v3 = (void*)v25 + sizeof(*v25);
+    }
+#if BETA_TEST
+	if (verbose>=5) printf("fetch: good return\n");
+#endif    
+    return buf_valid=0;
 }
 
 
+int purge(void)
+{
+    int i, fd;
+    int seek = PROBESEG*16;
+    
+    if (verbose>=6) printf("purge: called\n");
+#if 0
+    if (verbose>=6) {fetch(); dirty=1;}		/* test of checksumming */
+#endif
+    if (dirty <= 0) return 0;		/* nothing to purge */
+    
+    if ((i=fetch())) return i;	/* return error from fetch */
+    
+    i = buf.s5.length;
+    *(long*)buf.s5.checksum = crc32((char*)&buf.s5 + 4, i-4, CRC_POLY1);
+    
+    if ((fd=open("/dev/mem", O_WRONLY)) < 0) pdie("purge: can't open /dev/mem");
+    if (lseek(fd, seek, SEEK_SET) != seek) pdie("purge: ");
+    i = sizeof(buf.b);
+    if (write(fd, &buf.b, i) != i) pdie("purge: ");
+    close(fd);
+    
+    if (verbose>=6) printf("purge: successful write\n");
+    
+    return dirty = 0;			/* buffer is unmodified */
+}
+
+
+static int notice(int needed)
+{
+    if (fetch() || buf.s5.version < needed) {
+	printf(
+#if 0
+		"This information request requires that you previously booted your system\n"
+		"using LILO version %s or later.  These versions of the LILO boot\n"
+		"loader provide the BIOS data check information in low memory.  A boot floppy\n"
+		"constructed with 'mkrescue' may help.\n\n",
+#endif
+		"The information you requested is not available.\n\n"
+		"Booting your system with LILO version %s or later would provide the re-\n"
+		"quested information as part of the BIOS data check.  Please install a more\n"
+		"recent version of LILO on your hard disk, or create a bootable rescue floppy\n"
+		"with the 'mkrescue' command.\n\n",
+		
+		needed==4 ? "22.0" :
+		needed==5 ? "22.5.1" :
+		needed==6 ? "22.5.7" :
+		S(VERSION_MAJOR) "." S(VERSION_MINOR) VERSION_EDIT  );
+	return 1;
+    }
+    return 0;
+}
+
 
 /* print out the help page for -T flag */
 static void do_help(void)
@@ -216,7 +286,7 @@ static void do_help(void)
     printf("sizeof(video_t) = %d  sizeof(floppy_t) = %d  sizeof(hard_t) = %d\n"
            "sizeof(edd_t) = %d  sizeof(buf.s) = %d\n",
             sizeof(video_t),  sizeof(floppy_t), sizeof(hard_t), sizeof(edd_t),
-            sizeof(buf.s) );
+            sizeof(buf.s5) );
 	
     printf("fetch returns %d\n", fetch());            
 #endif
@@ -236,25 +306,29 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
     floppy_t *fd;
     int i;
     struct partition *pt_base;
+    unsigned long total;
 
+#if 0
     if((i=fetch())) {
         printf("No drive geometry information is available.\n\n");
         exit(0);
     }
-
+#else
+    if (notice(4)) exit(0);
+#endif
 #ifdef DEBUG
 	printf("get_geom: drive = 0x%02X\n", drive);
 	fflush(stdout);
 #endif
-    if (drive >= 0 && drive < buf.s.mflp) {
-	fd = (floppy_t*)&buf.b[buf.s.floppy] + drive;
+    if (drive >= 0 && drive < buf.s5.mflp) {
+	fd = (floppy_t*)&buf.b[buf.s5.floppy] + drive;
 	hd = (hard_t*)fd;
     }
     else if (drive == 0x80) {
-	hdp[drive-0x80] = hd = (hard_t*)&buf.b[buf.s.hard];
+	hdp[drive-0x80] = hd = (hard_t*)&buf.b[buf.s5.hard];
     }    
-    else if (drive >= 0x81 && drive < 0x80+buf.s.mhrd) {
-	if (drive > buf.s.disk) return 1;
+    else if (drive >= 0x81 && drive < 0x80+buf.s5.mhrd) {
+	if (drive > buf.s5.disk) return 1;
 	if (!hdp[drive-0x80]) {
 	    i = get_geom(drive-1, geom);
 #ifdef DEBUG
@@ -301,24 +375,28 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
    if (hd->fn08.flags&1 || hd->fn08.ah || hd->fn08.cx==0)
      return 1 + hd->fn08.ah;
    
+   i =
    geom->n_sect = hd->fn08.cx & 0x3F;
+   i *=
    geom->n_head = ((hd->fn08.dx>>8)&0xFF)+1;
+   i *=
    geom->n_cyl  = (((hd->fn08.cx>>8)&0xFF)|((hd->fn08.cx&0xC0)<<2))+1;
+   if (i > geom->n_total_blocks) geom->n_total_blocks = i;
    geom->n_disks = hd->fn08.dx & 0xFF;
    geom->pt = NULL;
 
    if (drive < 4)  return 0;
    
    pt_base = NULL;
-   if (buf.s.disk) {
-	pt_base = (struct partition *)&buf.b[buf.s.partitions];
+   if (buf.s5.disk) {
+	pt_base = (struct partition *)&buf.b[buf.s5.partitions];
    }
-   if (pt_base && drive <= (int)buf.s.disk) {
+   if (pt_base && drive <= (int)buf.s5.disk) {
 #if 0
    				geom->pt = &pt_base[(drive&15)*4];
 #else
 	void *p = (void*)pt_base;
-	int i = buf.s.version >= 4 ? 8 : 0;
+	int i = buf.s5.version >= 4 ? 8 : 0;
 	
 	p += (drive & 15) * (PART_TABLE_SIZE + i) + i;
 	geom->pt = (struct partition *)p;
@@ -342,7 +420,7 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
       geom->EDD_rev = hd->fn41.ah;
    }
    
-   if ((geom->EDD_flags) & EDD_SUBSET) {
+   if (((geom->EDD_flags) & EDD_SUBSET) || buf.s5.version >= 6) {
       edd_t *dp;
 
       dp = (edd_t*)hdp[drive-0x80 + 1];
@@ -360,27 +438,35 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
 	printf("get_geom: int13, fn=48\n");
 	fflush(stdout);
 #endif
-      
-      if ((dp->info) & EDD_PARAM_GEOM_VALID) {
-         if ((geom->n_sect != dp->sectors ||
-             geom->n_head != dp->heads) && !nowarn && !(warned[drive-0x80]&1) ) {
-                fprintf(errstd,"Warning: Int 0x13 function 8 and function 0x48 return different\n"
-                               "head/sector geometries for BIOS drive 0x%02X\n", drive);
-                show_geom("fn 08", geom->n_cyl, geom->n_head, geom->n_sect);
-                show_geom("fn 48", dp->cylinders, dp->heads, dp->sectors);
-		warned[drive-0x80] |= 1;
-	    }
 
-/* prefer to return the fn 8 geometry */
+	if ((dp->reg.flags&1) == 0 && dp->reg.ah == 0) {
+
+	      if ((dp->info) & EDD_PARAM_GEOM_VALID) {
+	         if ((geom->n_sect != dp->sectors || geom->n_head != dp->heads) &&
+				((verbose>0 && !lba32) || verbose>=4) && 
+				!nowarn && !(warned[drive-0x80]&1) ) {
+	                fprintf(errstd,"Warning: Int 0x13 function 8 and function 0x48 return different\n"
+                	               "head/sector geometries for BIOS drive 0x%02X\n", drive);
+        	        show_geom("fn 08", geom->n_cyl, geom->n_head, geom->n_sect);
+	                show_geom("fn 48", dp->cylinders, dp->heads, dp->sectors);
+			warned[drive-0x80] |= 1;
+		    }
+
+	/* prefer to return the fn 8 geometry */
 #if 0
-         geom->n_cyl  = dp->cylinders;
-         geom->n_head = dp->heads;
-         geom->n_sect = dp->sectors;
+	         geom->n_cyl  = dp->cylinders;
+	         geom->n_head = dp->heads;
+	         geom->n_sect = dp->sectors;
 #endif
-	 if (dp->total_sectors > geom->n_total_blocks)
-		geom->n_total_blocks = dp->total_sectors;
-      }
-
+		 total = dp->sectors;
+		 total *= dp->heads;
+		 total *= dp->cylinders;
+		 if (total > geom->n_total_blocks) geom->n_total_blocks = total;
+	
+	      }
+	      if (dp->total_sectors > geom->n_total_blocks)
+			geom->n_total_blocks = dp->total_sectors;
+	}
    }
    
    return 0;
@@ -390,11 +476,15 @@ static int get_geom(unsigned int drive, struct disk_geom *geom)
 /* get the conventional memory size in Kb */
 static int get_conv_mem(void)
 {
+#if 0
     if(fetch()) {
         printf("No memory information is available.\n\n");
         exit(0);
     }
-    return (int)buf.s.v.mem;
+#else
+    if (notice(4)) exit(0);
+#endif
+    return (int)eq->mem;
 }
 
 
@@ -410,19 +500,12 @@ static void do_ebda(void)
     printf("    Conventional Memory = %dK    0x%06X\n", m, m<<10);
     m <<= 10;
     m -= 0x200;
-#ifndef LCF_BUILTIN
-    n = m - (MAX_SECONDARY+4+MAX_DESCR_SECS)*SECTOR_SIZE;
-#else
     n = (select_loader()->size + SECTOR_SIZE - 1) / SECTOR_SIZE;
     n = m - (n+4+MAX_DESCR_SECS)*SECTOR_SIZE;
-#endif
     init = (n - (MAX_SETUPSECS+1)*SECTOR_SIZE)>>4;
     if (init > DEF_INITSEG) init = DEF_INITSEG;
-#if 0
-    if (!(verbose>0)) return;
-#else
     printf("\n");
-#endif
+
     printf("    The First stage loader boots at:  0x%08X  (0000:%04X)\n",
     			FIRSTSEG<<4, FIRSTSEG<<4);
     printf("    The Second stage loader runs at:  0x%08X  (%04X:%04X)\n",
@@ -432,17 +515,57 @@ static void do_ebda(void)
 
 }
 
+static char *number(unsigned long n)
+{
+    unsigned long k = 1000000000UL;	/* 10^9 */
+    static char s[16];
+    char *cp = s;
+    
+    while (n<k && k>1) k /= 1000UL;
+    sprintf(cp,"%lu",n/k);
+    n %= k;
+    k /= 1000UL;
+    while (k) {
+	while (*++cp) ;
+	sprintf(cp,",%03lu",n/k);
+	n %= k;
+	k /= 1000UL;
+    }
+    return s;
+}
+
 
 /* print the CHS geometry information for the specified disk */
 static void print_geom(int dr, struct disk_geom geom)
 {
-    char ch_ser[12] = { 0 };
+    char ch_ser[24] = { 0 };
+    char *sz = "KMGT";
+    char *sp = sz;
+    unsigned long n, m=0;
 
-    if (geom.serial_no) sprintf(ch_ser, "%08X", geom.serial_no);
+#ifdef DEBUG
+    if (!dr) geom.n_total_blocks = 4000000000UL;	/* 2.09Tb */
+#endif
+
+    if (geom.serial_no) sprintf(ch_ser, "vol-ID: %08X", geom.serial_no);
     printf("    bios=0x%02x, cylinders=%d, heads=%d, sectors=%d\t%s\n", 
 	dr, geom.n_cyl, geom.n_head, geom.n_sect, ch_ser);
-    if (geom.EDD_flags & EDD_PACKET)
-    printf("\tEDD packet calls allowed\n");
+    n = geom.n_total_blocks/2;
+    while (n > 999999) { n/=1000UL; n*=1024UL; n/=1000UL; sp++; }
+    if (n > 999) {
+	m = (n%1000UL)/10;
+	n /= 1000UL;
+	sp++;
+    }
+    if (m) printf("\t(%3lu.%02lu%cb", n, m, *sp);
+    else printf("\t(%3lu%cb", n, *sp);
+    
+    printf("%14s sectors)", number(geom.n_total_blocks));
+    if (geom.EDD_flags & EDD_PACKET) {
+/*	printf("\tEDD packet calls allowed");	*/
+	printf("\tLBA32 supported (EDD bios)");
+    } else printf("\tC:H:S supported (IBM PC bios)");
+    printf("\n");
 }
 
 
@@ -456,7 +579,7 @@ static void do_geom_all(void)
    for (d=0; d<16; d++) {
       dr = d+hd;
       if (get_geom(dr, &geom)==0) {
-         if (dr==0x80) printf("\nBIOS reports %d drive%s\n", (int) geom.n_disks,
+         if (dr==0x80) printf("\nBIOS reports %d hard drive%s\n", (int) geom.n_disks,
                                       (int)geom.n_disks==1 ? "" : "s");
       	 print_geom(dr, geom);
       }
@@ -470,9 +593,6 @@ static void do_geom(char *bios)
     int dr;
     struct disk_geom geom;
 
-#if 0    
-    dr = my_atoi(bios);
-#endif
     dr = to_number(bios);
     if (get_geom(dr, &geom)==0) print_geom(dr, geom);
     else printf("Unrecognized BIOS device code 0x%02x\n", dr);
@@ -526,50 +646,28 @@ static void print_pt(int index, struct partition pt)
 /* partition table display */
 static void do_table(char *part)
 {
-    int fd, i;
-    struct partition pt [PART_MAX];
-    unsigned int second, base;
-    unsigned short boot_sig;
+    struct partition pt [PART_MAX_MAX+1];
+    long volid;
+    long long where[PART_MAX_MAX+1];
+    int i,j;
+    int extd = (extended_pt || verbose>0);
     
-    if (!strncmp(part, "/dev/", 5)) {
-	if (strncmp(part+5, "md", 2) && isdigit(part[strlen(part)-1]) )
-	    die("Not a device: '%s'", part);
-	fd = open(part, O_RDONLY);
-    } else
-	fd = -1;
-    if (fd<0) die("Unable to open '%s'", part);
-    if (lseek(fd, PART_TABLE_OFFSET, SEEK_SET)<0) die("lseek failed");
-    if (read(fd, pt, sizeof(pt)) != sizeof(pt)) die("read pt failed");
-    if ( read(fd, &boot_sig, sizeof(boot_sig)) != sizeof(boot_sig)  ||
-	boot_sig != BOOT_SIGNATURE ) die("read boot signature failed");
-    {
-	if (lseek(fd, MAX_BOOT_SIZE+2, SEEK_SET)<0) die("lseek s/n failed");
-	if (read(fd, &second, sizeof(second)) != sizeof(second))
-	    die("read s/n failed");
-	printf(" S/N = %08X\n", second);
-    }
-    printf("%s\n", phead);
-    second=base=0;
+    j = read_partitions(part, extd ? nelem(pt)-1 : 0, &volid, pt, verbose>=5 ? where : NULL);
+    extd |= j>PART_MAX;
+    printf(" vol-ID: %08X\n\n%s\n", (int)volid, phead);
     for (i=0; i<PART_MAX; i++) {
 	print_pt(i+1, pt[i]);
-	if (is_extd_part(pt[i].sys_ind)) {
-	    if (!base) base = pt[i].start_sect;
-	    else die("invalid partition table: second extended partition found");
-	}
     }
-    i=5;
-    while (verbose>0 && base) {
-        if (llseek(fd, LLSECTORSIZE*(base+second) + PART_TABLE_OFFSET, SEEK_SET) < 0)
-            die("secondary llseek failed");
-	if (read(fd, pt, sizeof(pt)) != sizeof(pt)) die("secondary read pt failed");
-	if ( read(fd, &boot_sig, sizeof(boot_sig)) != sizeof(boot_sig)  ||
-	    boot_sig != BOOT_SIGNATURE ) die("read second boot signature failed");
-        print_pt(i++, pt[0]);
-        if (is_extd_part(pt[1].sys_ind)) second=pt[1].start_sect;
-        else base = 0;
+    i=4;
+    if (extd)
+    while (pt[i].sys_ind) {
+	print_pt(i+1, pt[i]);
+	i++;
+    }
+    if (verbose>=5) {
+    	printf("\n");
+    	for (i=0; i<j; i++) printf("%4d%20lld%12d\n", i+1, where[i], (int)(where[i]/SECTOR_SIZE));
     }
-        
-    close(fd);
 }
 
 
@@ -587,7 +685,6 @@ static void do_cr_pr(void)
     }
 }
 
-#if VIDEO
 static int egamem;
 static int mode, col, row, page;
 /*
@@ -604,15 +701,15 @@ int get_video(void)	/* return -1 on error, or adapter type [0..7] */
         
     if(fetch()) return -1;
 
-    if ((okay=buf.s.vid)) {
+    if ((okay=buf.s5.vid)) {
     /* get current video mode */
     /* reg.eax = 0x0F00;		*/
-	if (verbose >= 2)
+	if (verbose >= 6)
 	    printf("get video mode\n");
-	mode = buf.s.v.vid0F.al;
-	col = buf.s.v.vid0F.ah;
-	page = buf.s.v.vid0F.bh;
-	row = buf.s.v.vid0F.bl + 1;
+	mode = v1->vid0F.al;
+	col = v1->vid0F.ah;
+	page = v1->vid0F.bh;
+	row = v1->vid0F.bl + 1;
 	if (mode==7) {
 	    adapter=1;
 	    row = 25;
@@ -623,28 +720,26 @@ int get_video(void)	/* return -1 on error, or adapter type [0..7] */
 	} else adapter=2;	/* at least CGA */
     }
     if (okay>=2) {
-#if 1
     /* determine video adapter type */
     /*    reg.eax = 0x1200;	  call valid on EGA/VGA */
     /*    reg.ebx = 0xFF10;				*/
-	if (verbose >= 2)
+	if (verbose >= 6)
 	    printf("determine adapter type\n");
-	if ((unsigned)(monitor = buf.s.v.vid12.bh) <= 1U)  {
+	if ((unsigned)(monitor = v2->vid12.bh) <= 1U)  {
     	    adapter = 3; /* at least EGA */
-	    egamem = buf.s.v.vid12.bl;
+	    egamem = v2->vid12.bl;
 	}
 	else {
 	    okay = 0;
 	}
     }
-#endif
     if (okay>=2) {
 	/* check for VGA */
     	/* reg.eax = 0x1A00;	   get display combination */
-    	if (verbose >= 2)
+    	if (verbose >= 6)
 	    printf("get display combination\n");
-	if ( buf.s.v.vid1A.al==0x1A ) {
-    	    monitor = (buf.s.v.vid1A.bx >> ((verbose>=9)*8) ) & 0xFF;
+	if ( v2->vid1A.al==0x1A ) {
+    	    monitor = (v2->vid1A.bx >> ((verbose>=9)*8) ) & 0xFF;
     	    switch (monitor) {
     	      case 1:
     	        adapter = 1;
@@ -670,22 +765,36 @@ int get_video(void)	/* return -1 on error, or adapter type [0..7] */
     	}
     }
     if (okay>=3 && adapter==5) {
+	/* check for BIOS bug (trashing DX) */
+    	if (v25 && verbose >= 6)
+	    printf("check Enable Screen Refresh\n");
+	if (v25) {
+	    video_36_bug = 2;	/* mark implemented */
+	    
+	    if ((v25->vid36.ax & 0xFF) != 0x12)  video_36_bug = 1;
+	    else {
+		if (v25->vid36.cx != 0x1234 || v25->vid36.bp != 0x4321)
+		    video_36_bug |= 4;
+		if (v25->vid36.dx != 0x5680) video_36_bug |= 8;
+	    }
+	}
+	
 	/* check for VESA extensions */
-    	if (verbose >= 2)
+    	if (verbose >= 6)
 	    printf("check VESA present\n");
 	    
-	if ((buf.s.v.vid4F00.ax == 0x4f) && strncmp("VESA", buf.s.v.vid4F00.sig, 4)==0)  adapter++;
+	if ((v3->vid4F00.ax == 0x4f) && strncmp("VESA", v3->vid4F00.sig, 4)==0)  adapter++;
 	
 	if (adapter > 5) {
 	    /* reg.eax = 0x4f01;
 	       reg.ecx = 0x0101;	   640x480x256 */
-	    if ((buf.s.v.vid101.ax == 0x4f) && (buf.s.v.vid101.bits & 0x19) == 0x19) adapter ++;
+	    if ((v3->vid101.ax == 0x4f) && (v3->vid101.bits & 0x19) == 0x19) adapter ++;
 	    else adapter--;
 	}
 	if (adapter > 6) {
 	    /* reg.eax = 0x4f01;
 	       reg.ecx = 0x0103;	   800x600x256 */
-	    if ((buf.s.v.vid103.ax == 0x4f) && (buf.s.v.vid103.bits & 0x19) == 0x19) ;
+	    if ((v3->vid103.ax == 0x4f) && (v3->vid103.bits & 0x19) == 0x19) ;
 	    else adapter--;
 	}
     }
@@ -704,13 +813,15 @@ static char *display[] = { "unknown", "MDA", "CGA", "EGA",
     int adapter; /* 0=unknown, 1=MDA,HGC, 2=CGA, 3=EGA, 4=MCGA, 5=VGA,
     			 6=VESA (640), 7=VESA (800) */
 
+    if (notice(4)) exit(0);
+    
     adapter = get_video();
-
+#if 0
     if(adapter<0) {
         printf("No video mode information is available.\n");
         return;
     }
-    
+#endif
     printf("%s adapter:\n\n", display[adapter]);
     if (adapter < 3 || (adapter == 3 && egamem < 1) ) {
         printf("No graphic modes are supported\n");
@@ -726,9 +837,14 @@ static char *display[] = { "unknown", "MDA", "CGA", "EGA",
 	if (adapter >= 7)
 	    printf("    800x600x256   mode 0x0103\n");
     }
-
+    if (video_36_bug && (verbose>0 || (video_36_bug&(8+4)))) {
+     /* setting video_36_bug is a side effect of get_video */
+	printf("\nEnable Screen Refresh %s.\n",
+	    (video_36_bug & 4) ? "bugs are present" :
+	    (video_36_bug & 8) ? "bug is present" :
+	    (video_36_bug == 1) ? "is not supported" : "is supported");
+    }
 }
-#endif
 
 /* entry from lilo.c for the '-T' (tell) switch */
 void probe_tell (char *cmd)
@@ -761,7 +877,7 @@ int bios_max_devs(void)
     int i;
     
     if (!fetch() && !get_geom(0x80, &geom)) {
-	i = (buf.s.disk & 0x7f) + 1;
+	i = (buf.s5.disk & 0x7f) + 1;
 	if (geom.n_disks == i) return i;
     }
     return BIOS_MAX_DEVS;    
@@ -790,6 +906,7 @@ static void dump_pt(unsigned char *pt)
 int bios_device(GEOMETRY *geo, int device)
 {
     int bios1, bios, match, fd;
+    int bios2, snmatch;
     int mbios[BD_MAX_HARD];
     struct disk_geom bdata;
     DEVICE dev;
@@ -801,10 +918,14 @@ int bios_device(GEOMETRY *geo, int device)
     if (fetch()) return -1;
     
     if (verbose>=4) printf("bios_dev:  device %04X\n", device);
-    
+#ifdef DEBUG
+	fflush(stdout);
+#endif    
+    if (!has_partitions(device)) return -1;
+        
     match = 0;
     bios1 = -1;		/* signal error */
-    for (bios=0x80; bios<=buf.s.disk; bios++) {
+    for (bios=0x80; bios<=buf.s5.disk; bios++) {
 	mbios[bios-0x80] = 0;
 	if (get_geom(bios, &bdata)) break;
 	if (geo->cylinders == bdata.n_cyl &&
@@ -843,35 +964,45 @@ int bios_device(GEOMETRY *geo, int device)
 #endif
 
     if (verbose>=4) printf("bios_dev: geometry check found %d matches\n", match);
-    match = 0;
-    bios1 = -1;
+    
+    snmatch = match = 0;
+    bios2 = bios1 = -1;
+
+ /* 'bios' is set leaving the 'for' above */
     while (--bios >= 0x80) {
 	get_geom(bios, &bdata);
 	if (verbose>=4) {
-		printf("bios_dev: (0x%02X)  S/N=%08X  *PT=%08X\n",
+		printf("bios_dev: (0x%02X)  vol-ID=%08X  *PT=%08X\n",
 			bios, bdata.serial_no, (int)bdata.pt);
 #ifdef DEBUG
 		dump_pt((void*)bdata.pt);
 #endif
 	}
-	if (!memcmp(part,bdata.pt,sizeof(part)) &&
-		!(bdata.serial_no && serial!=bdata.serial_no)
-	   ) {
+	if ( !memcmp(part,bdata.pt,sizeof(part)) ) {
 	    match++;
 	    bios1 = bios;
 	}
+	if ( bdata.serial_no && serial==bdata.serial_no ) {
+	    snmatch++;
+	    bios2 = bios;
+	}
     }
-    if (verbose>=2) printf("bios_dev: PT match found %d match%s (0x%02X)\n",
+    if (verbose>=4) printf("bios_dev: PT match found %d match%s (0x%02X)\n",
+    		match, match==1 ? "" : "es", bios1&255);
+    if (match != 1) {
+	match = snmatch;
+	bios1 = bios2;
+	if (verbose>=4) printf("bios_dev: S/N match found %d match%s (0x%02X)\n",
     		match, match==1 ? "" : "es", bios1);
+    }
 
-    if (match != 1) bios1 = -1;
-    else {	/* match == 1 */
+    if (match == 1) {
 	get_geom(bios1, &bdata);
 	if (  (geo->sectors && geo->sectors!=bdata.n_sect)  ||
 	       (geo->heads && geo->heads!=bdata.n_head)  )  {
 	    unsigned long nblocks = geo->cylinders * geo->heads * geo->sectors;
 
-	    if (!nowarn && !(warned[bios1-0x80]&2) ) {
+	    if (verbose>0 && !nowarn && !(warned[bios1-0x80]&2) ) {
 		fprintf(errstd,"Warning: Kernel & BIOS return differing head/sector geometries for device 0x%02X\n", bios1);
 		show_geom("Kernel", geo->cylinders, geo->heads, geo->sectors);
 		show_geom("  BIOS", bdata.n_cyl, bdata.n_head, bdata.n_sect);
@@ -884,9 +1015,10 @@ int bios_device(GEOMETRY *geo, int device)
 	    geo->cylinders = nblocks / (bdata.n_head*bdata.n_sect);
 #endif
 	}
+	return (geo->device = bios1);
     }
 
-    return (geo->device = bios1);
+    return -1;
 }
 
 
@@ -897,3 +1029,134 @@ static void do_bitmap(char *file)
     printf("...<unimplemented>...\n");
 }
 #endif
+
+static unsigned char dl,dh;
+
+static int get_bios(void)
+{
+    if (fetch() || buf.s5.version<5) return -1;
+#if BETA_TEST
+	if (verbose>=5) printf("get_bios 1\n");
+#endif
+    dl = eq->boot_dx;
+#if BETA_TEST
+	if (verbose>=5) printf("get_bios 2\n");
+#endif
+    dh = eq->boot_dx >> 8;
+    
+    return eq->boot_dx;
+}
+
+
+void check_bios(void)
+{
+#if BETA_TEST
+	if (verbose>=5) printf("check_bios 1\n");
+#endif
+    if (bios_passes_dl == DL_NOT_SET) {
+	bios_passes_dl = DL_UNKNOWN;
+	
+	if (get_bios() < 0) return;
+#if BETA_TEST
+	if (verbose>=5) printf("check_bios 2\n");
+#endif
+	if (dl==0xFE) {
+	    if ( !((dh>=0x80 && dh<=DEV_MASK) || dh==0) )
+		bios_passes_dl = DL_BAD;
+	}
+	else if ( dl>=0x80 && dl<=DEV_MASK ) bios_passes_dl = DL_MAYBE;
+	else if ( dl > 0 ) bios_passes_dl = DL_BAD;
+    }
+    /* already set, leave alone */
+#if BETA_TEST
+	if (verbose>=5) printf("check_bios 3  bios_passes_dl=%d\n", (int)bios_passes_dl);
+#endif
+}
+
+
+void do_bios(void)
+{
+    int code=1;
+    
+static char *ord[] =
+{ "first", "second", "3rd", "4th", "5th", "6th", "7th", "8th",
+  "9th", "10th", "11th", "12th", "13th", "14th", "15th", "16th" };
+    
+    notice(5);
+    if (get_bios() < 0)
+	printf("No information available on the state of DL at boot.\n");
+    else
+    {
+	printf("BIOS provided boot device is 0x%02x  (DX=0x%04X).\n",
+		code=(dl==0xFE ? dh : dl), eq->boot_dx);
+	bios_passes_dl = DL_NOT_SET;
+	check_bios();
+    }
+    printf("\nUnless overridden, 'bios-passes-dl = %s' will be assumed.",
+	bios_passes_dl == DL_BAD ? "no" :
+	bios_passes_dl == DL_MAYBE ? "maybe" :
+	bios_passes_dl == DL_GOOD ? "yes" : "unknown" );
+    if (bios_passes_dl > DL_BAD) {
+	char *cp = NULL;
+	char *cn = NULL;
+
+	if (code==0) { cp="floppy"; cn=ord[0]; }
+	if (code>=0x80 && code<=0x8f) { cp="hard"; cn=ord[code&15]; }
+	if (cp)
+	printf("  If you\nactually booted from the %s %s drive, then this assumption is okay.",
+		cn, cp);
+#if 0
+	if (bios_passes_dl == DL_MAYBE)
+	    printf("\nIf the BIOS always gets DL set correctly, you might consider specifying\n"
+	           "  'bios_passes_dl = yes' or '-Z1'.\n");
+#endif
+    }
+    printf("\n");
+}
+
+#if VOLID
+static void do_volid(void)
+{
+    int bios, n, i, k, sv, nv, unique;
+    struct disk_geom geom;
+    unsigned long serial[MAX_BIOS_DEVICES];
+    int uniq[MAX_BIOS_DEVICES], valid[MAX_BIOS_DEVICES];
+
+    if (notice(4)) exit(0);
+    
+    printf("\n  BIOS     Volume ID\n\n");
+    unique = 1;
+    nv = n = 0;
+    for (bios=0x80; bios<0x80+MAX_BIOS_DEVICES && !get_geom(bios, &geom); bios++) {
+	int uq = 1;
+	i = bios - 0x80;
+#if 0
+	if (i==0 || i==2) geom.serial_no = 0;
+	if (i==4) geom.serial_no = serial[2];
+#endif
+
+	serial[i] = geom.serial_no;
+	valid[i] = sv = serial_valid(geom.serial_no, bios);
+	nv |= !sv;
+	if (sv) for (k=0; k<i; k++) uq &= (geom.serial_no != serial[k]);
+	uniq[i] = uq;
+	unique &= uq;
+	n++;
+
+	printf("  0x%02X     %08lX %s%s\n", i+0x80, serial[i],
+		uq ? "" : "*", sv ? "" : "-");
+    }
+
+    printf("\nVolume ID's are%s unique.\n", unique ? " all" : " NOT");
+    if (nv)
+	printf("   '-' marks an invalid Volume ID which will be automatically updated\n"
+		"\tthe next time  /sbin/lilo  is executed.\n");
+    if (!unique)
+	printf("   '*' marks a volume ID which is duplicated.  Duplicated ID's must be\n"
+		"\tresolved before installing a new boot loader.  The volume ID may\n"
+		"\tbe cleared using the '-z' and '-M' switches.\n"
+		);
+    printf("\n");
+}
+#endif
+
