@@ -1,6 +1,13 @@
 /* temp.c  -  Temporary file registry */
+/* 
+Copyright 1992-1998 Werner Almesberger.
+Copyright 1999-2003 John Coffman.
+All rights reserved.
 
-/* Copyright 1992-1995 Werner Almesberger. See file COPYING for details. */
+Licensed under the terms contained in the file 'COPYING' in the 
+source directory.
+
+*/
 
 
 #include <stdlib.h>
@@ -8,6 +15,7 @@
 #include <string.h>
 #include <errno.h>
 
+#include "lilo.h"
 #include "common.h"
 #include "temp.h"
 
